from sklearn.datasets import load_iris
from everywhereml.data import Dataset
from everywhereml.sklearn.ensemble import RandomForestClassifier


iris_dataset = Dataset.from_sklearn(load_iris(), name='Iris')
iris_classifier = RandomForestClassifier(n_estimators=10, max_depth=10)


import re
import json
import numpy as np
from tempfile import TemporaryDirectory
from subprocess import check_output


class BaseRuntime:
    """
    Actually run code in a given language to test it works as expected
    """
    def __init__(self):
        """
        Constructor
        """
        self.files = []

    def __str__(self):
        """
        Get runtime name
        :return: str
        """
        return f'Runtime(language={self.language}, dialect={self.dialect})'

    def __repr__(self):
        """
        Get runtime name
        :return: str
        """
        return str(self)

    @property
    def main_file(self):
        raise NotImplementedError('main_file')

    @property
    def language(self):
        raise NotImplementedError('language')

    @property
    def dialect(self):
        raise NotImplementedError('dialect')

    def add_main(self, contents):
        """

        :param contents:
        :return:
        """
        self.add_file(self.main_file, contents)

    def add_file(self, filename, contents):
        """
        Save file in project directory
        :param filename: str
        :param contents: str
        :return:
        """
        self.files.append((filename, contents))

    def output(self, tmp_folder=None):
        """
        Run program and get results back
        :return:
        """
        with TemporaryDirectory() as folder:
            # write files
            if tmp_folder is not None:
                folder = tmp_folder

            for filename, contents in self.files:
                with open('%s/%s' % (folder, filename), 'w', encoding='utf-8') as file:
                    file.write(contents)

            # exec runtime and get output
            commands = self.get_cmd()

            if isinstance(commands[0], str):
                commands = [commands]

            for cmd in commands:
                output = check_output(cmd, cwd=folder).decode('utf-8').strip()

            # format output as json
            if not output.startswith('['):
                output = '[%s]' % output

            output = re.sub(r'\]\s*\[', '],[', output)
            output = re.sub(r',\s*\]', ']', output)

            try:
                return np.asarray(json.loads(output), dtype=np.float)
            except json.decoder.JSONDecodeError:
                return None

    def get_cmd(self):
        """
        Get exec command
        :return: list
        """
        raise NotImplementedError('you MUST implement get_cmd()')


import re


class PortEntry:
    """
    Interact with port entries from the Arduino CLI
    """
    def __init__(self, entry: str):
        """

        :param entry:
        :return:
        """
        self.entry = entry.strip()
        self.port = ''
        self.description = ''

        match = re.search(r'^([-_/.a-zA-z0-9]+?)\s+(.+)$', self.entry)

        if match is not None:
            self.port, self.description = match.groups()

    def __str__(self) -> str:
        """

        :return:
        """
        return f'{self.port} ({self.description})'

    def __repr__(self) -> str:
        """

        :return:
        """
        return f'{self.port} ({self.description})'

    def exact_match(self, query: str, fqbn: str = None) -> bool:
        """
        Test if entry matches exactly query
        :param query:
        :param fqbn:
        :return:
        """
        return (self.port.lower() == query.lower()
                or self.description.lower() == query.lower()
                or (fqbn is not None and fqbn in self.description))

    def starts_with(self, query: str) -> bool:
        """
        Test if entry starts with query
        :param query:
        :return:
        """
        return (self.port.lower().startswith(query.lower())
                or self.description.lower().startswith(query.lower()))

    def ends_with(self, query: str) -> bool:
        """
        Test if entry ends with query
        :param query:
        :return:
        """
        return (self.port.lower().endswith(query.lower())
                or self.description.lower().endswith(query.lower()))

    def contains(self, query: str) -> bool:
        """
        Test if entry contains query
        :param query:
        :return:
        """
        return (query.lower() in self.port.lower()
                or query.lower() in self.description.lower())


import numpy as np
from math import pi
from tqdm import tqdm
from scipy.ndimage import convolve1d
from everywhereml.data import Dataset
from everywhereml.code_generators import GeneratesCode


class HOG(GeneratesCode):
    """
    (Approximated) Histogram of Oriented Gradients image feature extractor
    """
    def __init__(self, block_size: int = 8, bins: int = 9, cell_size: int = 3, approximate_atan2: bool = True, name: str = 'HOG'):
        """

        :param block_size: int
        :param bins: int
        :param cell_size: int
        :param approximate_atan2: bool
        :param name: str
        """
        self.name = name
        self.width = None
        self.height = None
        self.block_size = block_size
        self.cell_size = cell_size
        self.bins = bins
        self.approximate_atan2 = approximate_atan2
        self.num_outputs = None

    def __repr__(self):
        """
        Convert to string
        :return:
        """
        return str(self)

    def __str__(self):
        """
        Convert to string
        :return:
        """
        return f"HOG(block_size={self.block_size}, bins={self.bins}, cell_size={self.cell_size})"

    @property
    def working_size(self):
        """

        :return: int
        """
        return self.width * self.height

    @property
    def dtype(self):
        """
        Get dtype of operator
        :return:
        """
        return 'float'

    def fit(self, dataset):
        """
        Fit dataset
        :param dataset:
        :return:
        """
        self.width = dataset.width
        self.height = dataset.height
        hog = self.hog(dataset.images[0])
        self.num_outputs = len(hog)

    def transform(self, dataset):
        """
        Transform dataset
        :param dataset:
        :return:
        """
        features = [self.hog(im) for im in tqdm(dataset.images, desc='HOG')]
        feature_names = ['hog%d' % i for i in range(len(features[0]))]

        return Dataset(
            name=dataset.name,
            X=np.asarray(features, dtype=np.float),
            y=dataset.y,
            feature_names=feature_names,
            target_names=dataset.target_names
        )

    def hog(self, im):
        """
        Extract HOG features
        :param im:
        :return:
        """
        kernel = np.asarray([1, 0, -1], dtype=np.float)
        hog = []

        # compute bins
        for y in range(0, im.shape[0], self.block_size):
            for x in range(0, im.shape[1], self.block_size):
                block = im[y:y + self.block_size, x:x + self.block_size]

                if block.shape[0] != block.shape[1]:
                    continue

                gradient_y = convolve1d(block.T, kernel, mode='constant')[1:-1, 1:-1].T
                gradient_x = convolve1d(block, kernel, mode='constant')[1:-1, 1:-1]
                gradient = np.sqrt(gradient_y ** 2 + gradient_x ** 2)
                angle = np.abs(np.arctan2(gradient_y, gradient_x) * 180 / pi).astype(np.int)
                hist = np.zeros(self.bins)
                d = 180 // self.bins

                for a, g in zip(angle.flatten(), gradient.flatten()):
                    hist[min(self.bins - 1, a // d)] += g

                hog += hist.tolist()

        # normalize
        cell_length = self.bins * self.cell_size

        for i in range(0, len(hog), cell_length):
            cells = hog[i:i + cell_length]
            m = max(cells)
            m = max(m, 0.0001)

            for j, x in enumerate(cells):
                hog[i + j] /= m

        return np.asarray(hog)

    def get_template_data(self, dialect=None):
        """

        :param dialect:
        :return:
        """
        return {
            'width': self.width,
            'height': self.height,
            'block_size': self.block_size,
            'cell_size': self.cell_size,
            'bins': self.bins,
            'pi': pi,
            'num_outputs': self.num_outputs,
            'approximate_atan2': self.approximate_atan2
        }

    def get_template_data_cpp(self, dialect=None):
        """

        :param dialect:
        :return:
        """
        return {
            'dtype': self.dtype
        }


from everywhereml.tests.runtime.BaseRuntime import BaseRuntime


class CppRuntime(BaseRuntime):
    """
    C++ runtime
    """
    @property
    def language(self):
        return 'cpp'

    @property
    def dialect(self):
        return None

    @property
    def main_file(self):
        return 'main.cpp'

    def get_cmd(self):
        """

        :return:
        """
        return [
            ['g++', '-w', 'main.cpp', '-o', 'main.o'],
            ['./main.o']
        ]


from sklearn.base import clone
from sklearn.model_selection import cross_val_score
from everywhereml.data import Dataset
from everywhereml.code_generators import GeneratesCode


class SklearnBaseClassifier(GeneratesCode):
    """
    Base class for classifiers from the sklearn package
    """
    def __init__(self, **kwargs):
        """
        Proxy all calls to sklearn classifier
        :param kwargs:
        """
        self.sklearn.__init__(self, **kwargs)
        self.dataset = None
        self.y_pred = None

    def __call__(self, *args, **kwargs):
        """
        Proxy all calls to sklearn classifier
        """
        return self.sklearn(self, *args, **kwargs)

    def __getattr__(self, item):
        """
        Proxy all calls to sklearn classifier
        """
        return getattr(self.sklearn, item)

    def __repr__(self):
        """
        Convert to string
        :return:
        """
        return str(self)

    def __str__(self):
        """
        Convert to string
        :return:
        """
        attributes = [attr for attr in dir(self)
                      if not attr.startswith('_')
                      and not attr.endswith('_')
                      and not attr in ['dataset', 'y_pred']
                      and not callable(getattr(self, attr))]
        args = ', '.join(['%s=%s' % (attr, getattr(self, attr)) for attr in attributes])

        return f'{self.class_name}({args})'

    @property
    def sklearn(self):
        """
        Get sklearn native class
        :return: type
        """
        return [base for base in self.__class__.__bases__ if base.__module__.startswith('sklearn.')][0]

    @property
    def num_outputs(self):
        """
        Get number of classes
        :return:
        """
        return self.dataset.num_outputs if self.dataset is not None else 0

    def get_params(self, *args, **kwargs):
        """
        Proxy all calls to sklearn classifier
        :param args:
        :param kwargs:
        :return: dict
        """
        try:
            return self.sklearn.get_params(self, *args, **kwargs)
        except IndexError:
            return {}

    def clone(self):
        """
        Clone classifier
        :return:
        """
        return clone(self)

    def fit(self, X, y=None, **kwargs):
        """
        Fit data
        :param X:
        :param y:
        :param kwargs:
        """
        self.dataset = Dataset.from_XY(X, y)
        self.sklearn.fit(self, self.dataset.X, self.dataset.y, **kwargs)

        return self

    def predict(self, X, y=None, **kwargs):
        """
        Predict target
        :param X:
        :param y:
        :param kwargs:
        """
        dataset = Dataset.from_XY(X, y)
        self.y_pred = self.sklearn.predict(self, dataset.X, **kwargs)

        return self.y_pred

    def fit_predict(self, X, y=None, **kwargs):
        """
        Fit data and return predictions on that data
        :param X:
        :param y:
        :param kwargs:
        """
        self.fit(self, X, y)

        return self.predict(X)

    def score(self, X, y=None, **kwargs):
        """
        Get score of classifier on test data
        :param X:
        :param y:
        :param kwargs:
        :return:
        """
        dataset = Dataset.from_XY(X, y)

        return self.sklearn.score(self, dataset.X, dataset.y, **kwargs)

    def cross_val_score(self, X, y=None, **kwargs):
        """
        Get cross validate scores
        :param X:
        :param y:
        :param cv:
        :param kwargs:
        :return:
        """
        dataset = Dataset.from_XY(X, y)

        return cross_val_score(self, dataset.X, dataset.y, **kwargs)

    def get_template_data_cpp(self, dialect=None):
        """

        :param dialect:
        :return:
        """
        return {
            'input_dtype': 'float'
        }



import numpy as np
from os import mkdir
from os.path import isdir
from serial import Serial
from time import time
from tqdm import tqdm
from skimage.io import imsave
from everywhereml.data import Dataset, ImageDataset
from everywhereml.data.collect.ByteStream import ByteStream
from everywhereml.data.collect.BaseCollector import BaseCollector


class SerialCollector(BaseCollector):
    """
    Collect dataset from Serial port
    (mainly for Arduino projects)
    """
    def __init__(
            self,
            port,
            baud=9600,
            start_of_frame=None,
            end_of_frame=None,
            num_features=None,
            feature_names=None,
            delimiter=',',
    ):
        """

        :param port:
        :param baud:
        :param start_of_frame:
        :param num_features: int
        :param delimiter: str
        """
        if feature_names is not None and num_features == 0:
            num_features = len(feature_names)

        assert start_of_frame is None or isinstance(start_of_frame, str), 'start_of_frame MUST be a stirng'
        assert end_of_frame is None or isinstance(end_of_frame, str), 'end_of_frame MUST be a stirng'
        assert num_features is None or isinstance(num_features, int) and num_features > 0, 'num_features MUST be greater than 0'
        assert feature_names is None or num_features is None or len(feature_names) == num_features, 'feature_names count MUST match num_features'

        self.port = port
        self.baud = baud
        self.start_of_frame = start_of_frame
        self.end_of_frame = end_of_frame
        self.num_features = num_features
        self.feature_names = feature_names
        self.delimiter = delimiter

    def collect_many_classes(
            self,
            dataset_name,
            num_samples=0,
            duration=0,
            timeout=5,
            decode=None,
            **kwargs
    ):
        """
        Collect samples for many classes
        :param duration:
        :param dataset_name: str
        :param num_samples: int
        :param timeout: int timeout for serial
        :return:
        """
        assert decode is None or callable(decode), 'decode MUST be callable'

        target_names = []
        X = []
        y = []
        target_idx = 0

        for target_name in self.ask_target_name():
            with Serial(port=self.port, baudrate=self.baud, timeout=timeout) as serial:
                # discard first line to be sure new line is properly formatted
                serial.readline()
                serial.reset_input_buffer()

                if duration > 0:
                    class_data = self._collect_by_duration(serial=serial, duration=duration, decode=decode)
                else:
                    class_data = self._collect_by_samples(serial=serial, num_samples=num_samples, decode=decode)

            print('Captured %d samples' % len(class_data))

            if input('Is this class ok? (y|n) ').strip().lower() != 'y':
                continue

            target_names.append(target_name)
            X += class_data
            y += [target_idx] * len(class_data)
            target_idx += 1

        assert len(X) > 0, 'No data captured'

        feature_names = [f'f{i}' for i in range(len(X[0]))] if self.feature_names is None else self.feature_names

        return Dataset(
            name=dataset_name,
            X=np.asarray(X),
            y=np.asarray(y, dtype=np.int),
            feature_names=feature_names,
            target_names=target_names
        )

    def collect_many_images(
            self,
            dataset_name,
            base_folder,
            shape,
            num_samples=0,
            duration=0,
            max_fps=0,
            timeout=5,
            **kwargs
    ):
        """
        Collect images for many classes
        :param shape:
        :param max_fps:
        :param base_folder:
        :param duration:
        :param dataset_name: str
        :param num_samples: int
        :param timeout: int timeout for serial
        :return:
        """
        target_names = []
        image_length = shape[0] * shape[1]

        for target_name in self.ask_target_name():
            with Serial(port=self.port, baudrate=self.baud, timeout=timeout) as serial:
                # discard first line to be sure new line is properly formatted
                serial.readline()
                serial.reset_input_buffer()

                stream = ByteStream(
                    source=lambda: serial.read(1024),
                    start_of_frame=self.start_of_frame,
                    end_of_frame=self.end_of_frame,
                    max_fps=max_fps
                )

                if duration > 0:
                    images = list(stream.collect_duration(duration))
                else:
                    images = list(stream.collect_samples(num_samples))

                print(f'Collected {len(images)} images')

                if input('Is this class ok? (y|n) ').strip().lower() != 'y':
                    continue

                assert len(images) > 0, 'No image captured'

                if not isdir(base_folder):
                    mkdir(base_folder, 0o777)

                if not isdir(f'{base_folder}/{target_name}'):
                    mkdir(f'{base_folder}/{target_name}', 0o777)

                for i, image in enumerate(images):
                    arr = np.asarray([int(b) for b in image], dtype=np.uint8)

                    if len(arr) == image_length:
                        imsave(f'{base_folder}/{target_name}/{time()}.{i}.jpg', arr.reshape(shape))

                target_names.append(target_name)

        return ImageDataset.from_nested_folders(
            name=dataset_name,
            base_folder=base_folder,
            file_types=['jpg'],
            only_folders=target_names
        )

    def _collect_by_duration(self, serial, duration, decode=None):
        """
        Collect data for the given amount of time
        :param serial:
        :param duration: int
        :return:
        """
        records = []

        with tqdm(total=duration) as progress:
            last_tick = int(time())
            current_tick = last_tick
            end_time = current_tick + duration

            while current_tick <= end_time:
                current_tick = int(time())
                try:
                    line = serial.readline().decode('utf-8').strip()
                except UnicodeDecodeError:
                    continue

                if decode is not None:
                    try:
                        record = decode(line)
                    except:
                        continue
                else:
                    if not self._line_is_well_formatted(line):
                        continue

                    record = [float(x) for x in self._trim(line).split(self.delimiter)]

                records.append(record)

                if last_tick != current_tick:
                    progress.update(1)
                    last_tick = current_tick

        return records

    def _collect_by_samples(self, serial, num_samples, decode):
        """
        Collect the given number of samples
        :param serial:
        :param num_samples:
        :param decode:
        :return:
        """
        records = []

        while num_samples < 1:
            try:
                num_samples = int(input('How many samples do you want to record? '))
            except ValueError:
                pass

        with tqdm(total=num_samples) as progress:
            captured_samples = 0

            while captured_samples < num_samples:
                try:
                    line = serial.readline().decode('utf-8').strip()
                except UnicodeDecodeError:
                    continue

                if decode is not None:
                    try:
                        record = decode(line)
                    except:
                        continue
                else:
                    if not self._line_is_well_formatted(line):
                        continue

                    record = [float(x) for x in self._trim(line).split(self.delimiter)]

                records.append(record)
                captured_samples += 1
                progress.update(1)

        return records

    def _line_is_well_formatted(self, line):
        """
        Test if line matches expected format
        :param line:
        :return:
        """
        if len(line) == 0:
            return False

        if self.start_of_frame is not None and not line.startswith(self.start_of_frame):
            return False

        if self.num_features is not None and self._trim(line).count(self.delimiter) != self.num_features - 1:
            return False

        return True

    def _trim(self, line):
        """
        Remove start of frame from line
        :param line: str
        :return: str
        """
        if self.start_of_frame is not None and line.startswith(self.start_of_frame):
            return line.replace(self.start_of_frame, '').strip()

        return line.strip()


import re


def basic_python_prettify(python_code):
    """
    Apply basic Python prettifying
    :param python_code: str
    :return: str
    """
    blank_lines_re1 = re.compile(r'\n\s*\n([ \t]+)(if|else|while|return)')
    blank_lines_re2 = re.compile(r'\n\s*\n\s*\n([ \t]+)def ')
    blank_lines_re3 = re.compile(r'\n\s*\n\s*\n')

    for i in range(10):
        python_code = blank_lines_re1.sub(lambda m: f'\n{m.group(1)}{m.group(2)}', python_code)
        python_code = blank_lines_re2.sub(lambda m: f'\n\n{m.group(1)}def ', python_code)
        python_code = blank_lines_re3.sub('\n\n', python_code)

    return python_code.strip()


from everywhereml.code_generators.jinja.helpers import is_list


def c_shape(arr):
    """
    Convert array shape to C format
    :param arr:
    :return:
    """
    current = arr
    shapes = []

    while is_list(current):
        shapes.append(str(len(current)))
        current = current[0]

    return ']['.join(shapes)


def to_c_comment(x):
    """
    Convert string or list to C-style comment block
    :param x:
    :return:
    """
    if isinstance(x, str):
        x = x.split('\n')

    return '\n'.join([f' * {line}' for line in x])


try:
    from collections import Iterable
except ImportError:
    from collections.abc import Iterable


def is_list(x):
    """
    Check if argument is a list-like object
    :param x:
    :return: bool
    """
    return isinstance(x, Iterable) and not isinstance(x, str)



import seaborn as sns
from sklearn.feature_selection import SelectKBest
from umap import UMAP
import matplotlib.pyplot as plt


class DatasetPlotter:
    """
    Plot views of dataset
    """
    def __init__(self, dataset):
        """
        Constructor
        :param dataset:
        """
        self.dataset = dataset

    def features_pairplot(self, n=None, frac=None, k=None, **kwargs):
        """
        Plot pairplot of features
        :param n: int subsample
        :param frac: float subsample
        :param k: int feature selection
        :return:
        """
        df = self.dataset.df
        feature_columns = [c for c in df.columns if not c.startswith('target')]
        target_columns = [c for c in df.columns if c.startswith('target')]

        if kwargs.get('palette', None) is None and self.dataset.num_outputs < 10:
            kwargs.update(palette=sns.color_palette("tab10")[:self.dataset.num_outputs])

        if n is not None:
            df = df.sample(n=n)
        elif frac is not None:
            df = df.sample(frac=frac)

        X = df.drop(target_columns, axis=1)
        y = df[target_columns[0]]

        if k is None and len(df.columns) > 10:
            k = 6

        if k is not None:
            k_best = SelectKBest(k=k)
            k_best.fit(X, y)
            cols = k_best.get_support(indices=True)
            feature_columns = [column for i, column in enumerate(feature_columns) if i in cols]

        sns.pairplot(df, hue='target_name', vars=feature_columns, **kwargs)

    def umap(self, **kwargs):
        """
        Plot UMAP (https://umap-learn.readthedocs.io/en/latest/)
        :return:
        """
        kwargs = {
            **{'n_neighbors': 15, 'min_dist': 0.1},
            **kwargs
        }
        X = UMAP(n_components=2, **kwargs).fit_transform(self.dataset.X)

        ax = plt.figure().add_subplot()
        scatter = ax.scatter(X[:, 0], X[:, 1], c=self.dataset.y)
        ax.legend(*scatter.legend_elements(), title="Classes")
        ax.set_xlabel('Component #1')
        ax.set_ylabel('Component #2')
        ax.set_title(f'UMAP of {self.dataset.name}')
        plt.show()


from os.path import basename


class H:
    """
    Sketch header file
    """
    def __init__(self, filename: str, contents: str = ''):
        """

        :param filename:
        :param contents:
        """
        if contents == '':
            contents = filename
            filename = basename(filename)

        if contents.endswith('.h'):
            with open(contents, encoding='utf-8') as file:
                contents = file.read()

        self.filename = filename
        self.contents = contents



from everywhereml.preprocessing import Pipeline
from everywhereml.tests.BaseTestCase import BaseTestCase


class BasePreprocessingTestCase(BaseTestCase):
    """

    """

    def _test_cpp(self):
        """

        :return:
        """
        for dataset in self.get_datasets():
            for pipeline in self.get_pipelines():
                if not isinstance(pipeline, Pipeline):
                    steps = pipeline if isinstance(pipeline, list) else [pipeline]
                    pipeline = Pipeline(name='Test', steps=steps)

                self._test_pipeline_cpp(pipeline=pipeline, dataset=dataset)


from skimage.color import rgb2gray, yuv2rgb


class Grayscale:
    """
    Convert color images to grayscale
    """
    def __init__(self, pixformat: str):
        """
        Constructor
        :param pixformat:
        """
        assert pixformat in ('gray', 'rgb', 'rgb5656', 'rgb888', 'yuv422')

        self.pixformat = pixformat

    def transform(self, image):
        """

        :param image:
        :return:
        """
        if self.pixformat == 'gray':
            return image
        if self.pixformat in ('rgb', 'rgb565', 'rgb888'):
            return rgb2gray(image)
        if self.pixformat == 'yuv422':
            return rgb2gray(yuv2rgb(image))
        raise ValueError(f'Bad pixformat: {self.pixformat}')



from everywhereml.preprocessing.MinMaxScaler import MinMaxScaler
from everywhereml.preprocessing.Window import Window
from everywhereml.preprocessing.SpectralFeatures import SpectralFeatures
from everywhereml.preprocessing.Pipeline import Pipeline



import numpy as np
from everywhereml.data import Dataset
from everywhereml.preprocessing import MinMaxScaler
from everywhereml.tests.preprocessing.BasePreprocessingTestCase import BasePreprocessingTestCase


class MinMaxScalerTest(BasePreprocessingTestCase):
    """

    """
    def get_pipelines(self):
        return [
            MinMaxScaler(),
            MinMaxScaler(low=-1, high=1)
        ]

    def get_datasets(self):
        X = np.random.random((100, 100))
        y = np.random.randint(0, 2, 100)

        return [
            Dataset.from_XY(X, y=y),
            Dataset.from_XY(100 * X, y=y),
            Dataset.from_XY(100 * X - 50, y=y),
        ]

    def test_cpp(self):
        self._test_cpp()




import numpy as np
from everywhereml.data import ImageDataset
from everywhereml.preprocessing.image import LBP
from everywhereml.tests.preprocessing.BasePreprocessingTestCase import BasePreprocessingTestCase


class LBPTest(BasePreprocessingTestCase):
    """

    """
    def get_pipelines(self):
        return [
            LBP(r=1, spacing=1, eps=0),
            # LBP(r=1, spacing=2, eps=0),
            # LBP(r=1, spacing=1, eps=5),
        ]

    def get_datasets(self):
        return [
            ImageDataset(
                name='Test',
                images=np.random.randint(0, 255, size=(10, 100, 100)),
                labels=['test'] * 10
            )
        ]

    def test_cpp(self):
        self._test_cpp()




import re
from time import sleep
from os import makedirs
from os.path import join, abspath, basename, isdir
from everywhereml.arduino.Ino import Ino
from everywhereml.arduino.Cli import Cli, cli as cli_singleton
from everywhereml.arduino.SerialIO import SerialIO


class Sketch:
    """
    Write Arduino sketches from Python
    """
    def __init__(self, name: str, folder: str = '', board: str = None):
        """

        :param name: str
        :param folder: str
        :param board: str
        """
        if folder == ':system:':
            folder = join(cli_singleton.get_user_directory(), 'sketches')

        folder = folder.rstrip('/')

        self.name = name
        self.board = board
        self.folder = folder if basename(folder) == self.name else join(folder, self.name)
        self.fqbn = None
        self.port = None
        self.output = ''
        self.is_successful = True
        self.files = {}
        self.stats = {}

        if not isdir(self.folder):
            makedirs(self.folder, 0o777)

    def __add__(self, other):
        """

        :param other:
        :return:
        """
        if isinstance(other, list):
            for o in other:
                self += o

            return self

        filename = f'{self.basename}.ino' if isinstance(other, Ino) else other.filename
        self.files[filename] = {
            'contents': other.contents
        }

        with open(self.path_to(filename), 'w', encoding='utf-8') as file:
            file.write(other.contents)

        return self

    @property
    def basename(self):
        return basename(self.name)

    @property
    def path(self):
        """
        Get absolute path to project folder
        :return:
        """
        return abspath(self.folder)
    
    @property
    def serial(self):
        """
        Get Serial port I/O
        :return:
        """
        return SerialIO(port=self.port)

    def path_to(self, filename: str) -> str:
        """
        Convert relative path to absolute path
        :param filename:
        :return:
        """
        return abspath(join(self.folder, filename))

    def cat(self, filename: str = None):
        """
        Read file
        :param filename:
        :return:
        """
        if filename is None:
            filename = f'{self.name}.ino'

        return self.files[filename]['contents']

    def compile(self, cli : Cli = None, board: str = None, *args):
        """
        Compile sketch
        :param cli:
        :param board:
        :return: str|bool
        """
        self.is_successful = False
        cli = cli or cli_singleton
        self.fqbn = cli.find_fqbn(board or self.fqbn or self.board)
        self.stats.update(
            flash=0,
            flash_max=0,
            flash_percent=0,
            ram=0,
            ram_max=0,
            ram_percent=0)

        cli.exec('compile', '--verify', '--fqbn', self.fqbn, *args, cwd=self.path)
        self.output = cli.any_output
        self.stats['compile_time'] = cli.exec_time

        # parse compile log
        flash_pattern = r'Sketch uses (\d+) bytes(.+?Maximum is (\d+))?'
        memory_pattern = r'Global variables use (\d+)(.+?Maximum is (\d+))?'
        flash_match = re.search(flash_pattern, self.output.replace("\n", ""))
        memory_match = re.search(memory_pattern, self.output.replace("\n", ""))

        if flash_match is not None:
            self.stats['flash'] = int(flash_match.group(1))

            if len(flash_match.groups()) > 2:
                self.stats['flash_max'] = int(flash_match.group(3))
                self.stats['flash_percent'] = self.stats['flash'] / self.stats['flash_max']

        if memory_match is not None:
            self.stats['ram'] = int(memory_match.group(1))

            if len(memory_match.groups()) > 2:
                self.stats['ram_max'] = int(memory_match.group(3))
                self.stats['ram_percent'] = self.stats['ram'] / self.stats['ram_max']

        self.is_successful = self.stats['flash'] > 0

        return self

    def upload(self, port: str = None, cli : Cli = None, board: str = None, force: bool = False, return_success: bool = False, *args):
        """
        Upload sketch
        :param port:
        :param cli:
        :param board:
        :param force: if True, use the provided port as-is
        :param return_success:
        :param args:
        :return:
        """
        self.is_successful = False
        cli = cli or cli_singleton
        self.fqbn = cli.find_fqbn(board or self.fqbn or self.board)
        self.port = cli.find_port(port or self.port, fqbn=self.fqbn) if not force else port

        cli.exec('upload', '--verify', '--fqbn', self.fqbn, '--port', self.port, *args, cwd=self.path)

        self.output = cli.any_output
        self.is_successful = re.search(r'leaving...|success', self.output.lower()) is not None
        self.stats['upload_time'] = cli.exec_time
        # if you interact with the Serial port too early, you can get an error
        sleep(2)

        return self




import numpy as np
from everywhereml.code_generators import GeneratesCode


class MinMaxScaler(GeneratesCode):
    """
    See sklearn.preprocessing.MinMaxScaler
    """
    def __init__(self, low=0, high=1):
        """
        Constructor
        :param low:
        :param high:
        """
        self.dataset = None
        self.num_inputs = None
        self.min = None
        self.max = None
        self.low = low
        self.high = high

    def __repr__(self):
        """
        Convert to string
        :return:
        """
        return str(self)

    def __str__(self):
        """
        Convert to string
        :return:
        """
        return f"MinMaxScaler(low={self.low}, high={self.high})"

    def fit(self, dataset):
        """
        Fit dataset
        :param dataset:
        :return:
        """
        self.dataset = dataset
        self.num_inputs = dataset.num_inputs
        self.min = np.min(dataset.X, axis=0)
        self.max = np.max(dataset.X, axis=0)

    def transform(self, dataset):
        """
        Transform dataset
        :param dataset:
        :return:
        """
        X = (dataset.X - self.min) / (self.max - self.min) * (self.high - self.low) + self.low
        X[X < self.low] = self.low
        X[X > self.high] = self.high

        return dataset.replace(X=X)

    def get_template_data(self, dialect=None):
        """
        Get data for code generation
        :param dialect: str
        :return: dict
        """
        return {
            'num_inputs': self.num_inputs,
            'offset': self.min,
            'scale': np.nan_to_num(1 / (self.max - self.min)) * (self.high - self.low),
            'offset2': self.low,
            'low': self.low,
            'high': self.high
        }



import re


class BoardEntry:
    """
    Interact with board entries from the Arduino CLI
    """
    def __init__(self, entry: str):
        """

        :param entry:
        :return:
        """
        self.entry = entry.strip()
        self.name = ''
        self.fqbn = ''

        match = re.search(r'^(.+?)\s+([-a-zA-Z0-9_]+:[-a-zA-Z0-9_]+:[-a-zA-Z0-9_]+)', self.entry)

        if match is not None:
            self.name, self.fqbn = match.groups()

    def __str__(self) -> str:
        """

        :return:
        """
        return f'{self.name} ({self.fqbn})'

    def __repr__(self) -> str:
        """

        :return:
        """
        return f'{self.name} ({self.fqbn})'

    def exact_match(self, query: str) -> bool:
        """
        Test if entry matches exactly query
        :param query:
        :return:
        """
        return self.name.lower() == query.lower() or self.fqbn.lower() == query.lower()

    def starts_with(self, query: str) -> bool:
        """
        Test if entry starts with query
        :param query:
        :return:
        """
        return self.name.lower().startswith(query.lower()) or self.fqbn.lower().startswith(query.lower())

    def ends_with(self, query: str) -> bool:
        """
        Test if entry ends with query
        :param query:
        :return:
        """
        return self.name.lower().endswith(query.lower()) or self.fqbn.lower().endswith(query.lower())

    def contains(self, query: str) -> bool:
        """
        Test if entry contains query
        :param query:
        :return:
        """
        return query.lower() in self.name.lower() or query.lower() in self.fqbn.lower()


import numpy as np
from itertools import product
from scipy.stats import mode
from everywhereml.code_generators import GeneratesCode


class Window(GeneratesCode):
    """
    Sliding window
    """
    def __init__(self, length, shift=1):
        """
        :param length: int
        :param shift: int|float complement to overlap
        """
        assert isinstance(length, int) and length > 1, 'length MUST be > 1'
        assert isinstance(shift, int) and shift <= length or isinstance(shift, float) and 0 < shift <= 1, 'shift MUST be an integer <= length or a float in the range ]0, 1]'

        self.length = length
        self.shift = shift if isinstance(shift, int) else int(length * shift)
        self.num_inputs = None

    def __repr__(self):
        """
        Convert to string
        :return:
        """
        return str(self)

    def __str__(self):
        """
        Convert to string
        :return:
        """
        return f"Window(length={self.length}, shift={self.shift})"

    def fit(self, dataset):
        """
        Fit dataset
        :param dataset:
        :return:
        """
        # nothing to fit
        self.num_inputs = dataset.num_inputs

        return dataset

    def transform(self, dataset):
        """
        Transform
        """
        assert self.num_inputs is not None, 'Unfitted'

        idx = self.idx(dataset.X)
        X = dataset.X[idx]
        X = X.reshape((-1, self.num_inputs * self.length))
        y = np.asarray([mode(window)[0][0] for window in dataset.y[idx]])
        feature_names = [f'{feature_name}_{i}' for i, feature_name in product(range(self.length), dataset.feature_names)]

        return dataset.replace(X=X, y=y, feature_names=feature_names).annotate(window_length=self.length)

    def idx(self, X):
        """
        Get dense indices of X array
        :param X:
        :return: ndarray
        """
        w = np.arange(self.length)
        t = np.arange(len(X) - self.length + 1)
        idx = (w + t.reshape((-1, 1)))

        return idx[::self.shift]

    def get_template_data(self):
        """
        Get template data
        """
        return {
            'length': self.length,
            'shift': self.shift,
            'num_inputs': self.num_inputs
        }


import numpy as np
import pandas as pd
from pandas import DataFrame
from sklearn.model_selection import train_test_split
from everywhereml.plot.DatasetPlotter import DatasetPlotter


class Dataset:
    """
    Dataset data structure
    """
    @staticmethod
    def looks_like(dataset):
        """
        Test if argument looks like a dataset
        :param dataset:
        :return: bool
        """
        return hasattr(dataset, 'X') and hasattr(dataset, 'y') and hasattr(dataset, 'feature_names') and hasattr(dataset, 'target_names')

    @staticmethod
    def from_XY(X, y, feature_names=None, target_names=None, name=None):
        """
        Construct dataset from X and y arrays
        :param X:
        :param y:
        :param feature_names:
        :param target_names:
        :param name:
        :return:
        """
        if Dataset.looks_like(X):
            dataset = X
            y = dataset.y
            X = dataset.X
            feature_names = feature_names or dataset.feature_names
            target_names = target_names or dataset.target_names

        return Dataset(
            name=name or 'Dataset',
            feature_names=feature_names,
            target_names=target_names,
            X=np.asarray(X, dtype=float),
            y=np.asarray(y, dtype=int) if y is not None else np.zeros(len(X))
        )

    @staticmethod
    def from_sklearn(dataset, name=None):
        """
        Construct dataset from sklearn.datasets functions
        :param dataset:
        :param name:
        :return:
        """
        return Dataset(
            name=name or dataset.filename,
            feature_names=dataset.feature_names,
            target_names=dataset.target_names,
            X=dataset.data,
            y=dataset.target
        )

    @staticmethod
    def from_csv(filename, name, target_column='target', target_name_column=None, **kwargs):
        """
        Construct dataset from csv file
        :param filename:
        :param name:
        :param target_column:
        :param target_name_column:
        :param kwargs:
        :return:
        """
        return Dataset.from_pandas(
            pd.read_csv(filename, **kwargs),
            name=name,
            target_column=target_column,
            target_name_column=target_name_column
        )

    @staticmethod
    def from_pandas(df, name, target_column='target', target_name_column=None):
        """
        Construct dataset from pandas dataframe
        :param df:
        :param name:
        :param target_column:
        :param target_name_column:
        :return:
        """
        numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
        numeric_df = df.select_dtypes(include=numerics)
        feature_names = [c for c in numeric_df.columns if c != target_column and c != target_name_column]
        target_names = df[target_name_column].unique() if target_name_column is not None else None

        return Dataset(
            name=name,
            feature_names=feature_names,
            target_names=target_names,
            X=df[feature_names].to_numpy(),
            y=df[target_column].to_numpy()
        )

    def __init__(self, name, X, y, feature_names=None, target_names=None):
        """
        Constructor
        :param name:
        :param X:
        :param y:
        :param feature_names:
        :param target_names:
        """
        self.name = name
        self.df = None
        self.X = None
        self.y = None
        self.feature_names = None
        self.target_names = None
        self.annotations = []

        feature_names = [f for f in feature_names] if feature_names is not None else ['f_%d' % i for i in range(X.shape[1])]
        target_names = [n for n in target_names] if target_names is not None else ['target_%d' % i for i in range(len(set(y)))]

        self.replace(X=X, y=y, feature_names=feature_names, target_names=target_names)

    @property
    def distinct_targets(self):
        """
        Get distinct targets, sorted
        :return:
        """
        return sorted(list(set(self.y)))

    @property
    def num_inputs(self):
        """
        Get number of inputs
        :return: int
        """
        return len(self.feature_names)

    @property
    def num_outputs(self):
        """
        Get number of outputs
        :return:
        """
        return len(self.distinct_targets)

    @property
    def class_map(self):
        """
        Get mapping from target id to target name
        :return: dict
        """
        return {i: target_name for i, target_name in enumerate(self.target_names)}

    @property
    def plot(self):
        """
        Get instance of dataset plotter
        :return:
        """
        return DatasetPlotter(self)

    def describe(self):
        """
        Describe dataset
        :return:
        """
        return self.df.describe()

    def replace(self, X=None, y=None, feature_names=None, target_names=None, **kwargs):
        """
        Replace dataset properties
        :param X:
        :param y:
        :param feature_names:
        :param target_names:
        :return:
        """
        if X is not None:
            self.X = X

        if y is not None:
            self.y = y

        if feature_names is not None:
            self.feature_names = feature_names

        if target_names is not None:
            self.target_names = target_names

        self._assert_consistency()
        self._update_df()

        return self

    def clone(self, X=None, y=None, name=None):
        """
        Clone dataset
        :param X:
        :param y:
        :param name:
        :return:
        """
        if X is None:
            X = self.X

        if y is None:
            y = self.y

        return Dataset(
            name=name,
            X=X.copy(),
            y=y.copy(),
            target_names=self.target_names,
            feature_names=self.feature_names
        ).annotate(*self.annotations)

    def annotate(self, *args, **kwargs):
        """
        Add annotation to dataset
        :param args:
        :param kwargs:
        :return: self
        """
        if len(args) > 0:
            for arg in args:
                if isinstance(arg, dict):
                    self.annotate(**arg)

        if len(kwargs) > 0:
            self.annotations.append(kwargs)

        return self

    def get_annotations_by_attribute(self, attribute, **kwargs):
        """
        Get annotation that has the given attribute
        :param attribute: str
        :return: iterator
        """
        search_by_value = 'value' in kwargs
        value = kwargs.get('value')

        for annotation in self.annotations:
            if attribute in annotation:
                if not search_by_value or annotation.get(attribute) == value:
                    yield annotation.get(attribute)

        return None

    def drop_columns(self, columns: list):
        """
        Remove columns by index or name
        :param columns: list
        :return: Dataset
        """
        keep_columns = [(i, c) for i, c in enumerate(self.feature_names)
                        if i not in columns and c not in columns]
        keep_columns_indices = [i for i, c in keep_columns]

        return self.replace(
            X=self.X[:, keep_columns_indices],
            feature_names=[c for i, c in keep_columns]
        )

    def apply(self, pipeline):
        """
        Apply pipeline
        :param pipeline:
        :return:
        """
        dataset = pipeline.fit_transform(self)

        return self.replace(**dataset.__dict__)

    def split(self, train_size=None, test_size=None, **kwargs):
        """
        Split into train/test
        :param train_size:
        :param test_size:
        :return:
        """
        assert (train_size is None or 0 < train_size < 1) or (test_size is None or 0 < test_size < 1), 'train or test size MUST be in the range [0, 1] exclusive'

        X_train, X_test, y_train, y_test = train_test_split(self.X, self.y, train_size=train_size, test_size=test_size, **kwargs)

        return (
            self.clone(name='%s Train' % self.name, X=X_train, y=y_train),
            self.clone(name='%s Test' % self.name, X=X_test, y=y_test)
        )

    def _update_df(self):
        """
        Update internal DataFrame
        :return:
        """
        data = np.hstack((self.X, self.y.reshape((-1, 1))))
        columns = self.feature_names + ['target']

        self.df = DataFrame(data, columns=columns)
        self.df['target_name'] = [self.target_names[int(target)] for target in self.y]

    def _assert_consistency(self):
        """
        Assert properties are consistent
        :return:
        """
        assert self.X is not None, 'X CANNOT be NONE'
        assert self.y is not None, 'y CANNOT be NONE'
        assert len(self.X) == len(self.y), 'X and y MUST have the same length. X has shape %s, y has shape %s' % (str(self.X.shape), str(self.y.shape))
        assert len(self.feature_names) == self.X.shape[1], 'feature_names MUST be None or have the same length of X.shape[1] (%d given, %d expected)' % (len(self.feature_names), self.X.shape[1])
        assert len(self.target_names) == len(set(self.y)), 'target_names MUST be None or have the same length as the number of labels'




class Ino:
    """
    Sketch main file
    """
    def __init__(self, contents: str):
        """

        :param contents:
        """
        if contents.endswith('.ino'):
            with open(contents, encoding='utf-8') as file:
                contents = file.read()

        self.contents = contents



import re
import numpy as np

from everywhereml.code_generators import GeneratesCode


class SpectralFeatures(GeneratesCode):
    """
    Extract features from time-series
    """

    def __init__(self, order=2, eps=1e-3):
        """
        Constructor
        """
        assert order in (1, 2), 'order MUST be either 1 or 2'

        self.num_inputs = None
        self.window_length = None
        self.eps = eps
        self.order = order

    def __repr__(self):
        """
        Convert to string
        :return:
        """
        return str(self)

    def __str__(self):
        """
        Convert to string
        :return:
        """
        return f"SpectralFeatures(num_inputs={self.num_inputs}, window_length={self.window_length}, order={self.order}, num_features={len(self.feature_names)})"

    @property
    def first_order_feature_names(self):
        """
        Get names of first-order features
        :return:
        """
        return [
            'maximum',
            'minimum',
            'abs_maximum',
            'abs_minimum',
            'mean',
            'abs_energy',
            'mean_abs_change',
            'cid_ce'
        ]

    @property
    def second_order_feature_names(self):
        """
        Get names of second-order features
        :return:
        """
        return [
            'std',
            'var',
            'count_above_mean',
            'count_below_mean',
            'first_position_of_max',
            'first_position_of_min',
            'max_count',
            'min_count',
            'has_large_std',
            'skew',
            'kurtosis',
            'variation_coefficient',
        ]

    @property
    def feature_names(self):
        """
        Get names of used features
        :return: list
        """
        if self.order == 1:
            return self.first_order_feature_names
        return self.first_order_feature_names + self.second_order_feature_names

    def fit(self, dataset):
        """
        Fit dataset
        :param dataset:
        :return:
        """
        # this step is made to be used with a Window
        window_length = list(dataset.get_annotations_by_attribute('window_length'))

        assert len(window_length) > 0, 'SpectralFeatures MUST follow a Window'
        assert dataset.num_inputs % window_length[-1] == 0, 'window length MUST be a divider of dataset.num_inputs'

        self.window_length = window_length[-1]
        self.num_inputs = dataset.num_inputs // self.window_length

        return dataset

    def transform(self, dataset):
        """
        Transform dataset
        :param dataset:
        :return:
        """
        assert dataset.num_inputs % self.num_inputs == 0, 'self.num_inputs MUST be a divisor of dataset.num_inputs'

        spectral_features = None
        feature_names = []
        eps = self.eps

        for k in range(self.num_inputs):
            series = dataset.X[:, k::self.num_inputs]

            # first-order features
            maximum = series.max(axis=1).reshape((-1, 1))
            minimum = series.min(axis=1).reshape((-1, 1))
            abs_maximum = np.abs(maximum)
            abs_minimum = np.abs(minimum)
            mean = series.mean(axis=1).reshape((-1, 1))
            abs_energy = (series ** 2).mean(axis=1)
            mean_abs_change = np.abs(np.diff(series, axis=1)).mean(axis=1)
            cid_ce = (np.diff(series, axis=1) ** 2).mean(axis=1)

            spectral_features_k = [
                maximum,
                minimum,
                abs_maximum,
                abs_minimum,
                mean,
                abs_energy,
                mean_abs_change,
                cid_ce
            ]

            # second-order features
            if self.order == 2:
                ts_zero_mean = series - mean
                std = series.std(axis=1).reshape((-1, 1))
                var = ((series - mean) ** 2).mean(axis=1).reshape((-1, 1))
                count_above_mean = (ts_zero_mean > eps).sum(axis=1)
                count_below_mean = (ts_zero_mean < -eps).sum(axis=1)
                first_position_of_max = np.argmax(series, axis=1)
                first_position_of_min = np.argmin(series, axis=1)
                max_count = (series > (maximum - np.abs(maximum) * 0.02)).sum(axis=1)
                min_count = (series < (minimum + np.abs(minimum) * 0.02)).sum(axis=1)
                has_large_std = (std > 0.25 * (maximum - minimum))
                skew = np.where(var < eps, 0, ((series - mean) ** 3) / (var ** 1.5)).mean(axis=1)
                kurtosis = np.where(np.abs(var) < eps, 0, ((series - mean) ** 4) / (var ** 2)).mean(axis=1)
                variation_coefficient = np.where(mean < eps, 0, std / mean)

                spectral_features_k += [
                    std,
                    var,
                    count_above_mean,
                    count_below_mean,
                    first_position_of_max,
                    first_position_of_min,
                    max_count,
                    min_count,
                    has_large_std,
                    skew,
                    kurtosis,
                    variation_coefficient,
                ]

            spectral_features_k = np.hstack([feature.reshape((-1, 1)) for feature in spectral_features_k])
            spectral_features = spectral_features_k if spectral_features is None else np.hstack((spectral_features, spectral_features_k))
            # Window adds a _0 suffix to all features
            original_feature_name = re.sub(r'_0$', '', dataset.feature_names[k])
            feature_names += [f'{original_feature_name}_{feature_name}' for feature_name in self.feature_names]

        return dataset.replace(X=spectral_features, feature_names=feature_names)

    def get_template_data(self):
        """
        Get data for template
        :return: dict
        """
        return {
            'num_inputs': self.num_inputs,
            'window_length': self.window_length,
            'num_features': len(self.feature_names),
            'order': self.order,
            'eps': getattr(self, 'eps', 1e-3)
        }



from logging import info


class Selector:
    """
    Choose a board from a number of ones
    """
    def __init__(self, entries):
        """

        :param entries:
        """
        assert len(entries) > 0, 'list of entries is empty'

        self.entries = sorted(set(entries), key=lambda e: self.to_string(e))
        self._allow_none = False

    def log(self, objects_name: str):
        """
        Log choices
        :param objects_name:
        :return:
        """
        info(f"Found {len(self.entries)} {objects_name}")

        if len(self.entries) == 1:
            info(f"Choosing {self.to_string(self.entries[0])} automatically")

        return self

    def allow_none(self):
        """
        Allow user to exit from selection
        """
        self._allow_none = True

        return self

    def select(self):
        """
        Select one of the entries
        :return:
        """
        if len(self.entries) == 1:
            return self.to_value(self.entries[0])

        while True:
            print('Choose one of the following:')

            for i, entry in enumerate(self.entries):
                print(f' [{i + 1}] {self.to_string(entry)}')

            if self._allow_none:
                print("(0 to exit)")

            try:
                idx = int(input('> '))

                if self._allow_none and idx == 0:
                    return None

                if idx < 1:
                    continue

                return self.to_value(self.entries[idx - 1])
            except ValueError:
                pass
            except IndexError:
                pass

    def to_string(self, entry) -> str:
        """
        Convert entry to string
        """
        return entry[0] if isinstance(entry, tuple) else str(entry)

    def to_value(self, entry):
        """
        Convert entry to value
        """
        return entry[1] if isinstance(entry, tuple) else entry


import os.path
from os import makedirs
from everywhereml.code_generators.jinja.Jinja import Jinja
from everywhereml.code_generators.prettifiers.basic_python_prettifier import basic_python_prettify


class GeneratesCode:
    """
    Mixin with methods to export a given class to one of the supported languages
    """
    @property
    def class_name(self):
        """
        Get class name
        :return:
        """
        return type(self).__name__

    @property
    def package_name(self):
        """
        Get package name
        :return:
        """
        return '.'.join(self.__module__.__str__().split('.')[:-1])

    @property
    def template_folder(self):
        """
        Get base template folder
        :return:
        """
        return self.package_name.replace('.', '/')

    def get_template_data(self):
        """
        Abstract method to get template data
        :return:
        """
        raise NotImplementedError('get_template_data()')

    def to_cpp(self, class_name=None, namespace=None, instance_name=None, dialect=None, **kwargs):
        """
        Generate C++ code
        :param class_name: str
        :param namespace: str
        :param instance_name: str
        :param dialect: str
        :return:
        """
        return self.to_language(
            'cpp',
            class_name=class_name,
            namespace=namespace,
            instance_name=instance_name,
            dialect=dialect,
            **kwargs
        )

    def to_cpp_file(self, filename, **kwargs):
        """
        Save C++ code to file
        :param filename:
        :param kwargs:
        :return:
        """
        return self.to_file(filename, self.to_cpp(**kwargs))

    def to_arduino(self, **kwargs):
        """
        Generate C++/Arduino code
        :return:
        """
        return self.to_cpp(dialect='arduino', **kwargs)

    def to_arduino_file(self, filename, **kwargs):
        """
        Save C++/Arduino code to file
        :param filename:
        :param kwargs:
        :return:
        """
        return self.to_file(filename, self.to_arduino(**kwargs))

    def to_python(self, class_name=None, instance_name=None, dialect=None, **kwargs):
        """
        Generate Python code
        :param class_name: str
        :param instance_name: str
        :param dialect: str
        :return:
        """
        return basic_python_prettify(self.to_language(
            'py',
            class_name=class_name,
            instance_name=instance_name,
            dialect=dialect,
            **kwargs
        ))

    def to_micropython(self, **kwargs):
        """
        Generate Micropython code
        :return:
        """
        return self.to_python(dialect='micro', **kwargs)

    def to_python_file(self, filename, **kwargs):
        """
        Save Python code to file
        :param filename:
        :param kwargs:
        :return:
        """
        return self.to_file(filename, self.to_python(**kwargs))

    def to_micropython_file(self, filename, **kwargs):
        """
        Save Micropython code to file
        :param filename:
        :param kwargs:
        :return:
        """
        return self.to_python_file(filename, dialect='micro', **kwargs)

    def to_js(self, **kwargs):
        """

        :param kwargs:
        :return:
        """
        return self.to_language('js', **kwargs)

    def to_js_file(self, filename, **kwargs):
        """

        :param filename:
        :param kwargs:
        :return:
        """
        return self.to_file(filename, self.to_js(**kwargs))

    def to_language(self, language, dialect=None, **kwargs):
        """
        Generate code in given language
        :param language: str
        :param dialect: str
        :param language_data: dict
        :param kwargs:
        :return:
        """
        language_data_function = f'get_template_data_{language}'
        language_data = (getattr(self, language_data_function)(dialect=dialect)
                            if hasattr(self, language_data_function) else {})

        jinja = Jinja(self.template_folder, language=language, dialect=dialect)
        jinja.update(**self.get_template_data())
        jinja.update(**language_data)
        jinja.update(**kwargs)
        jinja.update(this=self, id=id(self), UUID=f'UUID{id(self)}', description=str(self).split('\n'))

        return jinja.render(self.class_name)

    def to_language_file(self, language, filename, **kwargs):
        """
        Save generated code to file
        :param language:
        :param file:
        :param kwargs:
        :return:
        """
        if os.path.dirname(filename):
            makedirs(os.path.dirname(filename), 0o777, exist_ok=True)

        with open(filename, 'w', encoding='utf-8') as file:
            contents = self.to_language(language=language, **kwargs)
            file.write(contents)

            return contents

    def to_file(self, filename, contents):
        """
        Save generated code to file
        :param language:
        :param contents:
        :return:
        """
        if os.path.dirname(filename):
            makedirs(os.path.dirname(filename), 0o777, exist_ok=True)

        with open(filename, 'w', encoding='utf-8') as file:
            file.write(contents)

            return contents



import numpy as np
from everywhereml.code_generators.jinja.Environment import Environment
from everywhereml.code_generators.jinja.FileSystemLoader import FileSystemLoader
from everywhereml.code_generators.jinja.filters import c_shape, to_c_array, to_c_comment, to_variable_name
from everywhereml.code_generators.jinja.filters import to_py_comment, to_py_list


class Jinja:
    """
    Render Jinja templates
    """
    def __init__(self, base_folder, language, dialect):
        """
        Constructor
        :param base_folder:
        :param language:
        :param dialect:
        """
        self.loader = FileSystemLoader(base_folder, language=language, dialect=dialect)
        self.env = Environment(loader=self.loader, extensions=['jinja2_workarounds.MultiLineInclude'])
        self.data = {}

        self.env.add_global('enumerate', enumerate)
        self.env.add_global('isinstance', isinstance)
        self.env.add_global('len', len)
        self.env.add_global('np', np)
        self.env.add_global('range', range)
        self.env.add_global('str', str)
        self.env.add_global('zip', zip)

        self.env.add_filter('c_shape', c_shape)
        self.env.add_filter('to_c_array', to_c_array)
        self.env.add_filter('to_c_comment', to_c_comment)

        self.env.add_filter('to_py_comment', to_py_comment)
        self.env.add_filter('to_py_list', to_py_list)

        self.env.add_filter('to_variable_name', to_variable_name)

    def update(self, **kwargs):
        """
        Update data
        :param kwargs:
        :return:
        """
        self.data.update(**kwargs)

    def render(self, template_name, template_data=None):
        """
        Return generated code
        :param template_name: str
        :param template_data: dict
        :return:
        """
        template_data = {
            **self.data,
            **(template_data or {})
        }
        return self.env.get_template(template_name).render(template_data)



import numpy as np
from math import log2
from itertools import product
from everywhereml.data import Dataset
from everywhereml.code_generators import GeneratesCode


class LBP(GeneratesCode):
    """
    Local Binary Pattern image feature extractor
    """
    def __init__(self, r=1, p=8, bins=64, spacing=1, eps=0):
        """

        :param r:
        :param p:
        :param bins:
        :param spacing: int compute LBP once every nth pixels (both along x and y axes)
        :param eps: int LBP returns 1 if (pixel - neighbor) > eps
        """
        assert r in (1, 2), 'r MUST be either 1 or 2'
        assert p == 8, 'p MUST be 8'
        assert bins and (not(bins & (bins - 1))), 'bins MUST be a power of 2'

        self.width = None
        self.height = None
        self.r = r
        self.p = p
        self.bins = bins
        self.spacing = spacing
        self.eps = eps
        self._lbp = None

    def __repr__(self):
        """
        Convert to string
        :return:
        """
        return str(self)

    def __str__(self):
        """
        Convert to string
        :return:
        """
        return f"LBP(r={self.r}, p={self.p}, spacing={self.spacing}, eps={self.eps})"

    @property
    def working_size(self):
        """

        :return: int
        """
        return self.width * self.height

    @property
    def bit_shift(self):
        """

        :return: int
        """
        return int(log2(256 // self.bins))

    @property
    def dtype(self):
        """
        Get dtype of operator
        :return:
        """
        return 'float'

    def fit(self, dataset):
        """
        Fit dataset
        :param dataset:
        :return:
        """
        self.width = dataset.width
        self.height = dataset.height

    def transform(self, dataset):
        """
        Transform dataset
        :param dataset:
        :return:
        """
        lbps = [self.lbp(im) for im in dataset.images]
        self._lbp = lbps
        hist = [self.hist(lbp) for lbp in lbps]
        feature_names = ['hist%d' % i for i in range(len(hist[0]))]

        return Dataset(
            name=dataset.name,
            X=np.asarray(hist, dtype=np.int),
            y=dataset.y,
            feature_names=feature_names,
            target_names=dataset.target_names
        )

    def lbp(self, im):
        """

        :param im:
        :return:
        """
        if im.max() <= 1:
            im = (im * 255).astype(np.uint8)

        if self.r == 1:
            roi = im[1:-1, 1:-1]
            roi = np.where(roi > 255 - self.eps, 255, roi + self.eps)
            d0 = np.where(im[0:-2, 0:-2] > roi, 1, 0).astype(np.uint8)
            d1 = np.where(im[0:-2, 1:-1] > roi, 1, 0).astype(np.uint8)
            d2 = np.where(im[0:-2, 2:] > roi, 1, 0).astype(np.uint8)
            d3 = np.where(im[1:-1, 2:] > roi, 1, 0).astype(np.uint8)
            d4 = np.where(im[2:, 2:] > roi, 1, 0).astype(np.uint8)
            d5 = np.where(im[2:, 1:-1] > roi, 1, 0).astype(np.uint8)
            d6 = np.where(im[2:, 0:-2] > roi, 1, 0).astype(np.uint8)
            d7 = np.where(im[1:-1, 0:-2] > roi, 1, 0).astype(np.uint8)
        elif self.r == 2:
            roi = im[2:-2, 2:-2] + self.eps
            roi = np.where(roi > 255, 255, roi).astype(np.uint8)
            d0 = np.where(im[1:-3, 1:-3] > roi, 1, 0).astype(np.uint8)
            d1 = np.where(im[0:-4, 2:-2] > roi, 1, 0).astype(np.uint8)
            d2 = np.where(im[1:-3, 3:-1] > roi, 1, 0).astype(np.uint8)
            d3 = np.where(im[2:-2, 4:] > roi, 1, 0).astype(np.uint8)
            d4 = np.where(im[3:-1, 3:-1] > roi, 1, 0).astype(np.uint8)
            d5 = np.where(im[4:, 2:-2] > roi, 1, 0).astype(np.uint8)
            d6 = np.where(im[3:-1, 1:-3] > roi, 1, 0).astype(np.uint8)
            d7 = np.where(im[2:-2, 0:-4] > roi, 1, 0).astype(np.uint8)
        else:
            raise AssertionError('LBP.r MUST be either 1 or 2')

        lbp = (
                d0
               | (d1 << 1)
               | (d2 << 2)
               | (d3 << 3)
               | (d4 << 4)
               | (d5 << 5)
               | (d6 << 6)
               | (d7 << 7)
        ).flatten()

        if self.spacing > 1:
            w = self.width - 2 * self.r
            xs = np.arange(w)[::self.spacing]
            ys = np.arange(self.height - 2 * self.r)[::self.spacing] * w
            idx = np.asarray([x + y for x, y in product(ys, xs)], dtype=np.int)
            lbp = lbp[idx]

        return lbp

    def hist(self, arr):
        """
        Compute histogram
        :param arr:
        :return:
        """
        return np.histogram(arr, bins=self.bins)[0]

    def get_template_data(self, dialect=None):
        """

        :param dialect:
        :return:
        """
        return {
            'width': self.width,
            'height': self.height,
            'r': self.r,
            'p': self.p,
            'bins': self.bins,
            'shift': self.bit_shift,
            'spacing': self.spacing,
            'eps': self.eps
        }

    def get_template_data_cpp(self, dialect=None):
        """

        :param dialect:
        :return:
        """
        return {
            'dtype': self.dtype
        }


import hexdump
import tensorflow as tf
from everywhereml.code_generators import GeneratesCode


class TensorFlowPorter(GeneratesCode):
    """
    Convert TF models to C++
    """
    def __init__(self, model, X, y):
        """

        :param model:
        """
        self.model = model
        self.num_inputs = X.shape[1] if len(X.shape) > 1 else 1

        if y.dtype == int:
            self.num_outputs = len(set(y)) if len(y.shape) == 1 or y.shape[1] == 1 else y.shape[1]
        else:
            # float ys -> regression
            self.num_outputs = 1

    def get_template_data(self):
        """

        :return:
        """
        return {
            'num_inputs': self.num_inputs,
            'num_outputs': self.num_outputs
        }

    def get_template_data_cpp(self, **kwargs):
        """

        :return:
        """
        converter = tf.lite.TFLiteConverter.from_keras_model(self.model)
        converted = converter.convert()
        bytes = hexdump.dump(converted).split(' ')
        bytes_array = ', '.join(['0x%02x' % int(byte, 16) for byte in bytes])

        return {
            'model_size': len(bytes),
            'bytes_array': bytes_array
        }


def tf_porter(*args, **kwargs):
    """
    Factory method
    :param kwargs:
    :return:
    """
    return TensorFlowPorter(*args, **kwargs)


from everywhereml.code_generators.jinja.filters.c_shape import c_shape
from everywhereml.code_generators.jinja.filters.to_c_array import to_c_array
from everywhereml.code_generators.jinja.filters.to_c_comment import to_c_comment
from everywhereml.code_generators.jinja.filters.to_py_comment import to_py_comment
from everywhereml.code_generators.jinja.filters.to_py_list import to_py_list
from everywhereml.code_generators.jinja.filters.to_variable_name import to_variable_name



from time import time


class ByteStream:
    """
    Read a binary stream
    """
    def __init__(self, source, start_of_frame, end_of_frame, max_fps=0):
        """

        :param source:
        :param start_of_frame:
        :param end_of_frame:
        :param max_fps:
        """
        assert callable(source), 'source MUST be callable'

        self.source = source
        self.start_of_frame = self.to_bytes(start_of_frame)
        self.end_of_frame = self.to_bytes(end_of_frame)
        self.max_fps = max_fps

    def collect_samples(self, count):
        """

        :param count:
        :return:
        """
        return self.collect(stop_condition=lambda c, d: c >= count)

    def collect_duration(self, duration):
        """

        :param duration:
        :return:
        """
        return self.collect(stop_condition=lambda c, d: d > duration)

    def collect(self, stop_condition):
        """

        :param duration: int
        :return:
        """
        assert callable(stop_condition), 'stop_condition MUST be callable'

        start = time()
        last_frame = start
        count = 0
        duration = 0
        buffer = b''

        while not stop_condition(count, duration):
            buffer += self.source()
            a = buffer.find(self.start_of_frame)
            b = buffer.find(self.end_of_frame)

            if a > b:
                buffer = buffer[a:]
                continue

            if a != -1 and b != -1:
                frame = buffer[a + len(self.start_of_frame):b]
                buffer = buffer[b + len(self.end_of_frame):]

                if self.max_fps == 0 or time() - last_frame >= 1.0 / self.max_fps:
                    yield frame
                    count += 1

            duration = time() - start

    def to_bytes(self, x):
        """

        :param x:
        :return:
        """
        if isinstance(x, str):
            return str.encode(x)
        if isinstance(x, bytes):
            return x
        raise AssertionError('x MUST either be a string or a byte array')


from everywhereml.code_generators import GeneratesCode


class Pipeline(GeneratesCode):
    """
    Run many pre-processing steps in sequence
    """
    def __init__(self, steps, name=''):
        """
        Constructor
        :param steps: list
        :param name: str
        """
        self.steps = steps
        self.name = name
        self.dataset = None
        self.num_inputs = 0
        self.num_outputs = 0
        self.working_size = 0

    def __repr__(self):
        """
        Convert to string
        :return:
        """
        return str(self)

    def __str__(self):
        """
        Convert to string
        :return:
        """
        steps = [' > %s' % str(step) for step in self.steps]

        return 'Pipeline:\n---------\n' + '\n'.join(steps)

    def __getitem__(self, item):
        """
        Get steps by index, slice or name
        :param item:
        :return: step or list of steps
        """
        if isinstance(item, int):
            return self.steps[item]

        if isinstance(item, str):
            return [step for step in self.steps if step.__class__.__name__ == item]

        if isinstance(item, slice):
            return self.steps[item]

        raise AssertionError('item MUST be an integer, a string or a slice')

    @property
    def dtype(self):
        """
        Get dtype of operators
        :return:
        """
        dtypes = [getattr(step, 'dtype', 'float') for step in self.steps]

        # if there's a single dtype, return
        if len(set(dtypes)) == len(dtypes):
            return dtypes[0]

        return 'float'

    def describe(self):
        """
        Convert to string
        :return:
        """
        return str(self)

    def fit(self, dataset):
        """
        Fit dataset
        :param dataset:
        :return:
        """
        self.num_inputs = dataset.num_inputs

        for step in self.steps:
            step.fit(dataset)
            dataset = step.transform(dataset)
            self.working_size = max(self.working_size, dataset.num_inputs, getattr(step, 'working_size', 0))

        self.dataset = dataset
        self.num_outputs = dataset.num_inputs

        return self

    def transform(self, dataset):
        """
        Transform dataset
        :param dataset:
        :return:
        """
        assert self.dataset is not None, 'Unfitted'

        for step in self.steps:
            dataset = step.transform(dataset)

        self.dataset = dataset

        return dataset

    def fit_transform(self, dataset):
        """
        Fit then transform dataset
        :param dataset:
        :return:
        """
        return self.fit(dataset).dataset

    def get_template_data(self):
        """
        Get data for code generation
        :return: dict
        """
        assert self.dataset is not None, 'Unfitted'

        return {
            'pipeline': self,
            'steps': self.steps,
            'dataset': self.dataset,
            'num_inputs': self.num_inputs,
            'num_outputs': self.num_outputs,
            'working_size': self.working_size
        }

    def get_template_data_cpp(self, dialect=None):
        """

        :param dialect:
        :return:
        """
        return {
            'dtype': self.dtype
        }


from time import time
from tqdm import tqdm
from serial import Serial, SerialException


class SerialIO:
    """
    Interact with the Serial port
    """
    def __init__(self, port, baud : int = 115200):
        """
        :param port:
        :param baud: int
        """
        self.port = port
        self.baud = baud

    def read_lines(self, limit: int = -1, timeout: int = 120, confirm: bool = True, **kwargs):
        """
        Read lines
        :param limit:
        :param timeout:
        :param confirm:
        :return: list
        """
        with Serial(self.port, baudrate=self.baud, timeout=10, **kwargs) as serial:
            lines = []
            last_update = 0
            current_update = 0

            if confirm:
                input("Serial port is open. Press [Enter] when you're ready to start capturing ")

            with tqdm(total=100) as progress:
                elapsed = 0
                started_at = time()

                while (limit < 0 or len(lines) < limit) and elapsed < timeout:
                    try:
                        lines.append(serial.readline().decode("utf-8").strip())

                        if limit > 0:
                            current_update = (100 * len(lines)) // limit
                        else:
                            current_update = (100 * elapsed) // timeout

                        progress.update(current_update - last_update)
                        last_update = current_update

                    except UnicodeDecodeError:
                        pass

                    finally:
                        elapsed = time() - started_at

            return lines[1:-1]


import numpy as np
import os.path
import random
from os import listdir
from skimage.io import imread
from skimage.color import rgb2gray
import matplotlib.pyplot as plt
from skimage import transform
from skimage.transform import resize, rotate, AffineTransform
from skimage.util import random_noise
from skimage.filters import gaussian
from skimage.exposure import adjust_log
from everywhereml.data import Dataset


class ImageDataset:
    """
    Dataset of images
    """
    @staticmethod
    def from_nested_folders(name, base_folder, file_types=None, skip_folders=None, only_folders=None, images_per_folder=0):
        """
        Load images from nested folders
        :param images_per_folder:
        :param only_folders:
        :param skip_folders:
        :param file_types:
        :param name:
        :param base_folder:
        :return:
        """
        if file_types is None:
            file_types = ['jpg', 'jpeg', 'png']

        # ignore file type case
        file_types = [ftype.lower() for ftype in file_types]

        images = []
        labels = []

        if skip_folders is None:
            skip_folders = []

        for child_folder in sorted(listdir(base_folder)):
            label = child_folder
            child_folder = f'{base_folder}/{child_folder}'

            if not os.path.isdir(child_folder):
                continue

            if label in skip_folders:
                continue

            if only_folders is not None and label not in only_folders:
                continue

            for i, filename in enumerate(sorted(listdir(child_folder))):
                if 0 < images_per_folder < i:
                    break

                image_name, ext = os.path.splitext(filename)

                if ext[1:].lower() not in file_types:
                    continue

                images.append(imread(f'{child_folder}/{filename}'))
                labels.append(label)

        return ImageDataset(name=name, images=images, labels=labels)

    @staticmethod
    def from_csv(filename: str, shape: tuple, name=None, **kwargs):
        """

        :param filename: str
        :param shape: tuple
        :param name: str
        :return:
        """
        if name is None:
            name = os.path.basename(filename)

        dataset = Dataset.from_csv(filename=filename, name=name, **kwargs)
        images = dataset.X.reshape((-1,) + shape)
        labels = [dataset.target_names[yi] for yi in dataset.y.astype(np.uint8)]

        return ImageDataset(
            name=name,
            images=images,
            labels=labels
        )

    def __init__(self, name, images, labels, **kwargs):
        """

        :param name:
        :param images:
        :param labels:
        """
        self.name = name
        self.images = np.asarray(images)
        self.labels = labels

    def __str__(self):
        """
        Describe dataset
        :return:
        """
        return f'ImageDataset[{self.name}](num_images={len(self.images)}, num_labels={len(self.target_names)}, labels={self.target_names})'

    @property
    def target_names(self):
        """
        Get unique labels list
        :return: list
        """
        target_names = []

        for label in self.labels:
            if label not in target_names:
                target_names.append(label)

        return target_names

    @property
    def num_classes(self):
        """
        Get number of classes
        :return: int
        """
        return len(self.target_names)

    @property
    def shape(self):
        """
        Get shape of images
        :return: tuple
        """
        return self.images[0].shape

    @property
    def width(self):
        """
        Get images' width
        :return: int
        """
        return self.shape[1]

    @property
    def height(self):
        """
        Get images' height
        :return: int
        """
        return self.shape[0]

    @property
    def depth(self):
        """
        Get images' depth
        :return: int
        """
        return self.shape[2] if len(self.shape) == 3 else 1

    @property
    def num_inputs(self):
        """
        Get number of inputs
        :return:
        """
        return self.width * self.height * self.depth

    @property
    def X(self):
        """

        :return:
        """
        return np.asarray(self.images, dtype=self.images[0].dtype).reshape((-1, self.num_inputs))

    @property
    def y(self):
        """

        :return:
        """
        inverse_classmap = {label: i for i, label in enumerate(self.target_names)}

        return np.asarray([inverse_classmap[label] for label in self.labels], dtype=np.uint8)

    def class_iterator(self):
        """

        :return:
        """
        for class_idx in range(self.num_classes):
            yield class_idx, [im for i, im in enumerate(self.images) if self.y[i] == class_idx]

    def clone(self, shallow=True, **kwargs):
        """

        :param shallow:
        :return:
        """
        kwargs = {
            **self.__dict__,
            **kwargs
        }

        if not shallow:
            kwargs['images'] = [np.copy(im) for im in kwargs['images']]

        return ImageDataset(**kwargs)

    def uint8(self):
        """
        Convert images to uint8
        :return:
        """
        if np.asarray(self.images).max() <= 1:
            images = [(im * 255).astype(np.uint8) for im in self.images]
        else:
            images = self.images

        return self.clone(images=images)

    def gray(self):
        """
        Convert images to grayscale
        :return:
        """
        if self.depth == 3:
            return self.clone(images=[rgb2gray(im) for im in self.images])

        return self

    def log_gradient(self):
        """

        :return:
        """
        log_images = [np.log(1 + im.astype(np.uint16)) for im in self.images]
        gradient_images = [im[2:, 1:-1] - im[0:-2, 1:-1] + im[1:-1, 2:] - im[1:-1, :-2] for im in log_images]

        return self.clone(images=gradient_images)

    def resize(self, new_shape, **kwargs):
        """
        Resize images to new shape
        :param new_shape:
        :return:
        """
        return self.clone(images=[resize(im, new_shape, **kwargs) for im in self.images])

    def map(self, fn):
        """
        Map images and labels
        :param fn:
        :return:
        """
        mapped = [fn(im, label) for im, label in zip(self.images, self.labels)]
        self.images = [m[0] if len(m) == 2 else m for m in mapped]
        self.labels = [m[1] if len(m) == 2 else label for m, label in zip(mapped, self.labels)]

        return self

    def apply(self, pipeline):
        """
        Apply pipeline
        :param pipeline:
        :return:
        """
        dataset = pipeline.fit_transform(self)

        return self.clone(**dataset.__dict__)

    def preview(self, samples_per_class, rows_per_class=1, figsize=(20, 10), **kwargs):
        """
        Show preview of images
        :param rows_per_class: int
        :param samples_per_class: int
        :param figsize:
        :return:
        """
        rows = self.num_classes * rows_per_class
        cols = samples_per_class // rows_per_class
        fig, ax = plt.subplots(rows, cols)
        fig.set_size_inches(figsize[0], figsize[1])

        for i, images in self.class_iterator():
            for j, k in enumerate(np.random.choice(len(images), samples_per_class)):
                row = j // cols
                col = j % cols
                ax[i * rows_per_class + row, col].imshow(images[k], **kwargs)

    def augment(
            self,
            augment_factor,
            scale_range=(1, 1.2),
            rotation_range=(-20, 20),
            translation_range=(-10, 10),
            illumination_range=(-0.1, 0.1),
            contrast_range=(0.95, 1.05),
            shear_range=(-0.05, 0.05),
            noise_range=(0.001, 0.002),
            blur_range=(0.9, 1.1)
    ):
        """
        Perform image augmentation
        :param translation_range:
        :param scale_range:
        :param blur_range:
        :param noise_range:
        :param shear_range:
        :param augment_factor:
        :param rotation_range:
        :param illumination_range:
        :param contrast_range:
        :return:
        """
        augmented_images = []
        augmented_labels = []
        target_names = self.target_names
        M = 255 if self.images.max() > 1 else 1

        def lin(r):
            return np.linspace(*r, 30)

        def cap(im):
            return np.where(im > 1, 1, np.where(im < 0, 0, im))

        def roi(im):
            h, w = self.images[0].shape[:2]
            H, W = im.shape[:2]
            return im[H // 2 - h // 2:H // 2 - h // 2 + h, W // 2 - w // 2:W // 2 - w // 2 + W]

        available_transforms = [
            [lambda im: roi(resize(im, (im.shape[0] * zoom, im.shape[1] * zoom))) for zoom in lin(scale_range)] if scale_range is not None else None,
            [lambda im: rotate(im, angle, mode='edge') for angle in lin(rotation_range)] if rotation_range is not None else None,
            [lambda im: transform.warp(im, AffineTransform(translation=(t, 0)), mode='edge') for t in lin(translation_range)] if translation_range is not None else None,
            [lambda im: transform.warp(im, AffineTransform(translation=(0, t)), mode='edge') for t in lin(translation_range)] if translation_range is not None else None,
            [lambda im: cap(im + illu) for illu in lin(illumination_range)] if illumination_range is not None else None,
            [lambda im: adjust_log(im, gain) for gain in lin(contrast_range)] if contrast_range is not None else None,
            [lambda im: transform.warp(im, AffineTransform(shear=shear), mode='edge') for shear in lin(shear_range)] if shear_range is not None else None,
            [lambda im: random_noise(im, var=noise) for noise in lin(noise_range)] if noise_range is not None else None,
            [lambda im: gaussian(im, sigma=sigma) for sigma in lin(blur_range)] if blur_range is not None else None,
        ]
        available_transforms = [t for t in available_transforms if t is not None]

        for class_idx, images in self.class_iterator():
            images_to_generate = int(len(images) * augment_factor)

            while images_to_generate > 0:
                im = random.choice(images)
                num_transformations_to_apply = random.randint(1, len(available_transforms))
                transform_indices = np.random.choice(np.arange(len(available_transforms)), size=num_transformations_to_apply, replace=False)
                transforms = [random.choice(available_transforms[i]) for i in transform_indices]

                if M > 1:
                    im = im.astype(np.float) / 255

                for t in transforms:
                    im = t(im)

                if im.shape != self.images[0].shape:
                    continue

                if M >= 1:
                    im = (im * 255).astype(np.uint8)

                augmented_images.append(im)
                augmented_labels.append(target_names[class_idx])
                images_to_generate -= 1

        self.augmented_images = augmented_images
        augmented_images = np.asarray(augmented_images, dtype=self.images.dtype).reshape((-1,) + self.images.shape[1:])

        return self.clone(images=np.vstack((self.images, augmented_images)), labels=self.labels + augmented_labels)






import json
import numpy as np
from tempfile import NamedTemporaryFile
from xgboost import XGBClassifier as Impl
from everywhereml.sklearn.SklearnBaseClassifier import SklearnBaseClassifier


class XGBClassifier(SklearnBaseClassifier, Impl):
    """
    xgboost.XGBClassifier wrapper
    """
    def __init__(self,
                 max_depth=None,
                 learning_rate=None,
                 n_estimators=100,
                 objective=None,
                 gamma=None,
                 min_child_weight=None,
                 max_delta_step=None,
                 subsample=None,
                 colsample_bytree=None,
                 colsample_bylevel=None,
                 colsample_bynode=None,
                 reg_alpha=None,
                 reg_lambda=None,
                 scale_pos_weight=None,
                 base_score=None,
                 random_state=None,
                 missing=np.nan,
                 num_parallel_tree=None,
                 monotone_constraints=None,
                 interaction_constraints=None,
                 importance_type="gain",
                 gpu_id=None,
                 validate_parameters=None,
                 **kwargs):
        """
        Patch constructor
        """
        super(XGBClassifier, self).__init__(
            max_depth=max_depth,
            learning_rate=learning_rate,
            n_estimators=n_estimators,
            objective=objective,
            gamma=gamma,
            min_child_weight=min_child_weight,
            max_delta_step=max_delta_step,
            subsample=subsample,
            colsample_bytree=colsample_bytree,
            colsample_bylevel=colsample_bylevel,
            colsample_bynode=colsample_bynode,
            reg_alpha=reg_alpha,
            reg_lambda=reg_lambda,
            scale_pos_weight=scale_pos_weight,
            base_score=base_score,
            random_state=random_state,
            missing=missing,
            num_parallel_tree=num_parallel_tree,
            monotone_constraints=monotone_constraints,
            interaction_constraints=interaction_constraints,
            importance_type=importance_type,
            gpu_id=gpu_id,
            validate_parameters=validate_parameters,
            **kwargs)

    @property
    def sklearn(self):
        """
        Get sklearn native class
        :return: type
        """
        return [base for base in self.__class__.__bases__ if base.__module__.startswith('xgboost.')][0]

    def get_template_data(self):
        """
        Get additional data for template
        :return: dict
        """
        with NamedTemporaryFile('w+', suffix='.json', encoding='utf-8') as tmp:
            self.save_model(tmp.name)
            tmp.seek(0)
            decoded = json.load(tmp)
            trees = decoded['learner']['gradient_booster']['model']['trees']

            return {
                'n_classes': int(decoded['learner']['learner_model_param']['num_class']) or 2,
                'trees': [{
                    'left': tree['left_children'],
                    'right': tree['right_children'],
                    'features': tree['split_indices'],
                    'thresholds': tree['split_conditions'],
                } for tree in trees]
            }



import numpy as np
from everywhereml.code_generators.jinja.helpers import is_list


def to_c_array(arr, precision=11, with_parentheses=True):
    """
    Convert array to C format
    :param arr: list|numpy.ndarray
    :param precision: int
    :param with_parentheses: bool
    :return:
    """
    arr = np.nan_to_num(arr)

    if not is_list(arr):
        fmt = '%%.%dff' % precision
        return fmt % arr

    contents = (', '.join([to_c_array(x, precision) for x in arr]))

    return '{%s}' % contents if with_parentheses else contents



from os import getcwd
from os.path import abspath
from platform import system
from timeit import default_timer
from logging import debug
from subprocess import STDOUT, CalledProcessError, check_output
from everywhereml.arduino.BoardEntry import BoardEntry
from everywhereml.arduino.PortEntry import PortEntry
from everywhereml.dialog.Selector import Selector


class Cli:
    """
    Arduino-CLI wrapper
    """
    def __init__(self, cwd: str = None):
        """

        """
        self.cwd = None
        self.exe = None
        self.output = None
        self.error = None
        self.exec_time = 0

        self.configure_exe('arduino-cli.exe' if system().lower() == 'window' else 'arduino-cli')
        self.set_working_dir(cwd)

    @property
    def is_successful(self) -> bool:
        """
        Test if command was successful
        @return bool True on success, False on error
        """
        return self.error is None

    @property
    def any_output(self):
        """
        Return either output or error
        :return: str
        """
        return self.output or self.error or ''

    @property
    def lines(self) -> list:
        """
        Get command output as lines
        @return list
        """
        return [line.strip() for line in (self.output or "").split("\n")]

    def configure_exe(self, exe: str):
        """
        Set executable path
        :param exe: str
        :return:
        """
        self.exe = exe

    def set_working_dir(self, cwd: str):
        """
        Set current working directory
        :param cwd: str
        :return:
        """
        self.cwd = abspath(cwd or getcwd())

    def find_fqbn(self, query: str):
        """
        Find fqbn from fuzzy query
        :param query: str
        :return:
        """
        assert query is not None, 'query cannot be None'

        boards = [BoardEntry(b.strip()) for b in self.exec('board', 'listall').split('\n') if b.strip()]
        exact_matches = [b for b in boards if b.exact_match(query)]

        if len(exact_matches) > 0:
            return Selector(exact_matches).select().fqbn

        partial_matches = (
                [b for b in boards if b.starts_with(query)] +
                [b for b in boards if b.ends_with(query)] +
                [b for b in boards if b.contains(query)])

        if len(partial_matches) > 0:
            return Selector(partial_matches).select().fqbn

        raise IndexError(f'Cannot find board that matches query "{query}"')

    def find_port(self, query: str = None, fqbn: str = None):
        """
        Find port by fuzzy query
        :param query:
        :param fqbn:
        :return:
        """
        ports = [PortEntry(p.strip()) for p in self.exec('board', 'list').split('\n') if p.strip()]

        # exact match by port query
        if query is not None:
            exact_matches = [p for p in ports if p.exact_match(query)]

            if len(exact_matches) > 0:
                return Selector(exact_matches).log("Serial ports").select().port

        # exact match by fqbn
        if fqbn is not None:
            exact_matches = [p for p in ports if p.exact_match("", fqbn)]

            if len(exact_matches) > 0:
                return Selector(exact_matches).log("Serial ports").select().port

        partial_matches = (
            [p for p in ports if p.starts_with(query)] +
            [p for p in ports if p.ends_with(query)] +
            [p for p in ports if p.contains(query)])

        if len(partial_matches) > 0:
            return Selector(partial_matches).log("Serial ports").select().port

        return query

    def exec(self, command: str, *args, cwd: str = None) -> str:
        """
        Execute command
        :return: str command output
        """
        try:
            debug(" ".join([self.exe, command] + list(args)))
            start_time = default_timer()
            self.output = check_output([self.exe, command] + list(args), stderr=STDOUT, cwd=(cwd or self.cwd)).decode('utf-8')
            self.error = None
            self.exec_time = default_timer() - start_time

            return self.output
        except CalledProcessError as err:
            self.error = err.output.decode('utf-8')
            self.output = None

            return self.error

        return ''

    def get_directory(self, which: str) -> str:
        """
        Get system directory

        :param which:
        :return: str
        """
        try:
            config = self.exec('config', 'dump')
            directories = config.split('directories:')[1]

            return [line.strip() for line in directories.split('\n') if line.strip().startswith(f'{which}:')][0][len(which)+1:].strip()
        except IndexError:
            return 'error'

    def get_user_directory(self) -> str:
        """
        Get system user directory
        :return:
        """
        return self.get_directory('user')


cli = Cli()


from sklearn.ensemble import RandomForestClassifier as Impl
from everywhereml.sklearn.SklearnBaseClassifier import SklearnBaseClassifier


class RandomForestClassifier(SklearnBaseClassifier, Impl):
    """
    sklearn.ensemble.RandomForestClassifier wrapper
    """
    def get_template_data(self):
        """
        Get additional data for template
        :return: dict
        """
        return {
            'trees': [{
                'left': tree.tree_.children_left,
                'right': tree.tree_.children_right,
                'features': tree.tree_.feature,
                'thresholds': tree.tree_.threshold,
                'classes': tree.tree_.value,
            } for tree in self.estimators_]
        }



import logging
import numpy as np
from unittest import TestCase
from numpy.testing import assert_allclose
from everywhereml.data import Dataset
from everywhereml.sklearn.ensemble import RandomForestClassifier
from everywhereml.code_generators.jinja.Jinja import Jinja
from everywhereml.tests.runtime.CppRuntime import CppRuntime


class RandomForestClassifierTest(TestCase):
    """

    """
    def setUp(self) -> None:
        """

        :return:
        """
        self.logger = logging.getLogger("default")
        self.datasets = [
            Dataset.from_XY(X=np.random.random((100, 100)), y=np.random.randint(0, 2, 100)),
            Dataset.from_XY(X=np.random.random((1000, 100)), y=np.random.randint(0, 5, 1000))
        ]
        self.instances = [
            RandomForestClassifier(n_estimators=20, max_depth=20)
        ]

    def test_cpp(self):
        runtime = CppRuntime()

        for dataset in self.datasets:
            for clf in self.instances:
                self.logger.info(f"Classifier: {clf}")

                clf.fit(dataset)
                y_true = clf.predict(dataset)
                main = Jinja('tests', language='cpp', dialect=None).render('classifier', {
                    'X': dataset.X,
                    'classifier': clf.to_cpp(instance_name='classifier')
                })
                runtime.add_file('main.cpp', main)
                output = runtime.output(tmp_folder='/Users/simone/Desktop/tmp')
                self.assertIsNotNone(output, "Output is None")

                assert_allclose(y_true, output, rtol=1e-2)



from everywhereml.preprocessing.image import HOG
from everywhereml.preprocessing.image.object_detection.BaseObjectDetectionPipeline import BaseObjectDetectionPipeline


class HogPipeline(BaseObjectDetectionPipeline):
    """
    Classify objects by Histogram of Oriented Gradients (HOG)
    """
    def __init__(self, hog_params: dict = None, **kwargs):
        """

        :param hog_params:
        :param kwargs:
        """
        kwargs.setdefault('name', 'HogPipeline')
        BaseObjectDetectionPipeline.__init__(self, **kwargs)
        self.steps = [HOG(**(hog_params or {}))]



from sklearn.tree import DecisionTreeClassifier as Impl
from everywhereml.sklearn.SklearnBaseClassifier import SklearnBaseClassifier


class DecisionTreeClassifier(SklearnBaseClassifier, Impl):
    """
    sklearn.ensemble.DecisionTreeClassifier wrapper
    """
    def get_template_data(self):
        """
        Get additional data for template
        :return: dict
        """
        return {
            'left': self.tree_.children_left,
            'right': self.tree_.children_right,
            'features': self.tree_.feature,
            'thresholds': self.tree_.threshold,
            'classes': self.tree_.value,
            'i': 0
        }



import os.path
from jinja2 import Environment as Base


class Environment(Base):
    """
    Custom Jinja environment
    """
    def join_path(self, template, parent):
        """
        Enable relative template paths
        :param template: str
        :param parent: str
        """
        return os.path.normpath(os.path.join(os.path.dirname(parent), template))

    def add_filter(self, name, function):
        """
        Add filter
        :param name: str
        :param function: callable
        :return:
        """
        assert callable(function), 'filter function MUST be callable'

        self.filters[name] = function

    def add_global(self, name, value):
        """
        Add global
        :param name: str
        :param value:
        :return:
        """
        self.globals[name] = value



import requests
from os import mkdir
from os.path import isdir, join, abspath
from time import time, sleep
from tqdm.auto import tqdm
from logging import info, debug, warning
from everywhereml.data import ImageDataset


class MjpegCollector:
    """
    Collect frames from MJPEG HTTP stream
    """
    def __init__(self, address, max_fps=0):
        """

        :param address:
        :param max_fps:
        """
        self.address = address
        self.max_fps = max_fps

    def collect_many_classes(self, dataset_name, base_folder, samples_per_class=0, duration=0, preview=False):
        """

        :param dataset_name:
        :param base_folder:
        :param num_samples:
        :param duration:
        :param preview:
        :return:
        """
        assert samples_per_class > 0 or duration > 0, 'you MUST set either samples_per_class or duration'

        print('This is an interactive data capturing procedure.')
        print('Keep in mind that when you enter a class name, the capturing will start in 2 seconds, so be ready!')

        while True:
            try:
                target_name = input('Which class are you going to capture? (leave empty to exit) ').strip()

                if len(target_name) == 0 and input('Are you sure you want to exit? (y|n) ').strip().lower() == 'y':
                    break

                if len(target_name) == 0:
                    continue

                # allow some time for the user to settle
                sleep(2)

                if duration > 0:
                    images = self.collect_by_duration(duration)
                else:
                    images = self.collect_by_samples(samples_per_class)

                info('Captured %d images' % len(images))

                if input('Is this class ok? (y|n) ').strip().lower()[0] != 'y':
                    continue

                # save to disk
                if not isdir(base_folder):
                    info(f'creating {abspath(base_folder)} folder')
                    mkdir(base_folder)

                # make sure folder is created
                while not isdir(base_folder):
                    info(f'Folder {abspath(base_folder)} does not exists and cannot be created. Please create manually')
                    input('Press [Enter] when you\'re done')

                target_folder = abspath(join(base_folder, target_name))

                if not target_folder:
                    info(f'creating {target_folder} folder')
                    mkdir(target_folder)

                # make sure folder is created
                while not isdir(target_folder):
                    info(f'Folder {target_folder} does not exists and cannot be created. Please create manually')
                    input('Press [Enter] when you\'re done')

                timestamp = time()

                for i, im in enumerate(images):
                    image_path = join(target_folder, f'{timestamp}_{i}.jpg')

                    with open(image_path, 'wb') as file:
                        debug(f'Saving image to {image_path}')
                        file.write(im)
            except Exception as ex:
                warning(ex)

        return ImageDataset.from_nested_folders(
            name=dataset_name,
            base_folder=base_folder,
            file_types=['jpg']
        )

    def collect_by_duration(self, duration):
        """
        Collect for a given duration
        :param duration:
        :return:
        """
        return self._collect(progress=lambda _i, _duration: _duration / duration)

    def collect_by_samples(self, num_samples):
        """
        Collect a given number of samples
        :param num_samples:
        :return:
        """
        return self._collect(progress=lambda _i, _duration: _i / num_samples)

    def _collect(self, progress):
        """
        Collect until progress reaches 1
        :param progress: callable
        :return:
        """
        bytes = b''
        count = 0
        start = time()
        last_frame = start
        last_progress = 0
        frames = []

        with tqdm(total=100) as progress_bar:
            with requests.get(self.address, stream=True) as res:
                for data in res.iter_content(1024):
                    bytes += data
                    a = bytes.find(b'\xff\xd8')
                    b = bytes.find(b'\xff\xd9')

                    if a != -1 and b != -1:
                        debug("detected new jpeg frame")
                        jpg = bytes[a:b + 2]
                        bytes = bytes[b + 2:]

                        if self.max_fps == 0 or time() - last_frame >= 1.0 / self.max_fps:
                            debug("appending to dataset")
                            frames.append(jpg)
                            count += 1
                            last_frame = time()

                    current_progress = progress(count, time() - start)
                    progress_bar.update(100 * (current_progress - last_progress))
                    last_progress = current_progress

                    if current_progress >= 1:
                        break

        return frames



import logging
import numpy as np
from unittest import TestCase
from numpy.testing import assert_allclose
from everywhereml.code_generators.jinja.Jinja import Jinja
from everywhereml.tests.runtime.CppRuntime import CppRuntime


class BaseTestCase(TestCase):
    def setUp(self):
        """

        :return:
        """
        self.logger = logging.getLogger("default")

    def _test_pipeline(self, runtime, pipeline, dataset, **kwargs):
        """

        :param runtime:
        :param pipeline:
        :param dataset:
        :param language:
        :param kwargs:
        :return:
        """
        self.logger.debug(f"Testing runtime={runtime}, pipeline={pipeline}, dataset={dataset}")

        dataset_t = pipeline.fit_transform(dataset.clone())
        main = Jinja('tests', language=runtime.language, dialect=runtime.dialect).render('pipeline', {
            'X': dataset.X,
            'pipeline': pipeline.to_cpp(instance_name='pipeline')
        })
        runtime.add_main(main)
        output = runtime.output(tmp_folder='/Users/simone/Desktop/tmp')
        self.assertIsNotNone(output, "Output is None")

        for x_true, x_pred in zip(dataset_t.X, output):
            #self.assertLess(abs((sum(x_true) - sum(x_pred)) / len(x_true)), 0.1, 'sum of values does not match')
            is_not_close = np.abs(x_true - x_pred) > 0.1
            self.assertLess(sum(is_not_close), len(is_not_close) * 0.01, f'{self.summarize_array(x_pred)} vs {self.summarize_array(x_true)}')

    def _test_pipeline_cpp(self, pipeline, dataset):
        """

        :return:
        """
        self._test_pipeline(
            runtime=CppRuntime(),
            pipeline=pipeline,
            dataset=dataset
        )

    def summarize_array(self, arr):
        """

        :param arr:
        :return:
        """
        start = ', '.join(['%.3f' % x for x in arr[:10]])
        end = ', '.join(['%.3f' % x for x in arr[-min(len(arr) - 10, 10):]])
        ellipsis = ' ... ' if len(arr) > 20 else ''

        return f'[{start}{ellipsis}{end}]'


from distutils.core import setup

packages=["everywhereml", "everywhereml.plot", "everywhereml.xgboost", "everywhereml.tests", "everywhereml.tests.xgboost", "everywhereml.tests.runtime", "everywhereml.tests.sklearn", "everywhereml.tests.sklearn.ensemble", "everywhereml.tests.templates", "everywhereml.tests.preprocessing", "everywhereml.tests.preprocessing.image", "everywhereml.tests.data", "everywhereml.dialog", "everywhereml.arduino", "everywhereml.sklearn", "everywhereml.sklearn.tree", "everywhereml.sklearn.ensemble", "everywhereml.get_started", "everywhereml.code_generators", "everywhereml.code_generators.jinja", "everywhereml.code_generators.jinja.filters", "everywhereml.code_generators.prettifiers", "everywhereml.templates", "everywhereml.templates.py", "everywhereml.templates.js", "everywhereml.templates.xgboost", "everywhereml.templates.xgboost.js", "everywhereml.templates.xgboost.cpp", "everywhereml.templates.sklearn", "everywhereml.templates.sklearn.tree", "everywhereml.templates.sklearn.tree.cpp", "everywhereml.templates.sklearn.ensemble", "everywhereml.templates.sklearn.ensemble.py", "everywhereml.templates.sklearn.ensemble.py.micro", "everywhereml.templates.sklearn.ensemble.cpp", "everywhereml.templates.cpp", "everywhereml.templates.preprocessing", "everywhereml.templates.preprocessing.image", "everywhereml.templates.preprocessing.image.cpp", "everywhereml.templates.preprocessing.image.object_detection", "everywhereml.templates.preprocessing.image.object_detection.cpp", "everywhereml.templates.preprocessing.image.transform", "everywhereml.templates.preprocessing.image.transform.cpp", "everywhereml.templates.preprocessing.cpp", "everywhereml.preprocessing", "everywhereml.preprocessing.image", "everywhereml.preprocessing.image.object_detection", "everywhereml.preprocessing.image.transform", "everywhereml.data", "everywhereml.data.collect"]
data=["tests/templates/pipeline.cpp.jinja", "tests/templates/classifier.cpp.jinja", "templates/py/vote.py.jinja", "templates/py/latency.py.jinja", "templates/py/class_map.py.jinja", "templates/py/BaseClassifier.py.jinja", "templates/js/BaseClassifier.js.jinja", "templates/js/class_map.js.jinja", "templates/js/vote.js.jinja", "templates/xgboost/js/tree.jinja", "templates/xgboost/js/XGBClassifier.jinja", "templates/xgboost/cpp/tree.jinja", "templates/xgboost/cpp/XGBClassifier.jinja", "templates/sklearn/tree/cpp/tree.cpp.jinja", "templates/sklearn/tree/cpp/DecisionTreeClassifier.cpp.jinja", "templates/sklearn/ensemble/py/micro/RandomForestClassifier.py.micro.jinja", "templates/sklearn/ensemble/py/micro/tree.py.jinja", "templates/sklearn/ensemble/cpp/tree.cpp.jinja", "templates/sklearn/ensemble/cpp/RandomForestClassifier.cpp.jinja", "templates/cpp/latency.cpp.jinja", "templates/cpp/vote.cpp.jinja", "templates/cpp/TensorFlowPorter.cpp.jinja", "templates/cpp/class_map.cpp.jinja", "templates/cpp/BaseClassifier.cpp.jinja", "templates/preprocessing/image/cpp/BaseImageStep.cpp.jinja", "templates/preprocessing/image/cpp/HOG.cpp.jinja", "templates/preprocessing/image/cpp/LBP.cpp.jinja", "templates/preprocessing/image/object_detection/cpp/HogPipeline.cpp.jinja", "templates/preprocessing/image/object_detection/cpp/BaseObjectDetectionPipeline.cpp.jinja", "templates/preprocessing/image/transform/cpp/Resize.cpp.jinja", "templates/preprocessing/cpp/Step.cpp.jinja", "templates/preprocessing/cpp/Pipeline.cpp.jinja", "templates/preprocessing/cpp/SpectralFeatures.cpp.jinja", "templates/preprocessing/cpp/Window.cpp.jinja", "templates/preprocessing/cpp/MinMaxScaler.cpp.jinja"]

setup(
  name='everywhereml',
  packages=packages,
  version='0.2.19',
  license='MIT',
  description='Train ML in Python, run everywhere',
  author='Simone Salerno',
  author_email='support@eloquentarduino.com',
  url='https://github.com/eloquentarduino/everywhereml',
  download_url='https://github.com/eloquentarduino/everywhereml/blob/master/dist/everywhereml-0.2.19.tar.gz?raw=true',
  keywords=[
    'ML',
    'machine learning'
  ],
  install_requires=[
    'numpy',
    'pandas',
    'seaborn',
    'scikit-learn',
    'scikit-image',
    'Jinja2',
    'cached-property',
    'umap-learn',
    'python-slugify',
    'hexdump',
    'jinja2_workarounds',
    'requests',
    'pySerial',
    'tqdm'
  ],
  extras_require={
    'tf': ['tensorflow']
  },
  package_data={
    'everywhereml': data
  },
  classifiers=[
    'Development Status :: 2 - Pre-Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Code Generators',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)


from slugify import slugify


def to_variable_name(name: str) -> str:
    """
    Convert name to well formed variable name
    :param name:
    :return:
    """
    return slugify(name).replace('-', '_')



from skimage.transform import resize
from everywhereml.code_generators import GeneratesCode


class Resize(GeneratesCode):
    """
    Resize image
    """
    def __init__(self, width: int, height: int, pixformat: str = None):
        """
        Constructor
        :param width:
        :param height:
        :param pixformat:
        """
        assert width > 0, 'width MUST be greater than 0'
        assert height > 0, 'height MUST be greater than 0'
        assert pixformat is None or pixformat in ('gray', 'rgb565', 'rgb888', 'yuv422'), 'pixformat MUST be one of (gray, rgb565, rgb888, yuv422)'

        # relative width
        if width < 1 < height:
            width *= height

        # relative height
        if height < 1 < width:
            height *= width

        assert width > 1 and height > 1, 'width and height CANNOT be both relative'

        self.width = width
        self.height = height
        self.pixformat = None
        self.input_shape = None

    def __repr__(self):
        """

        :return:
        """
        return str(self)

    def __str__(self):
        """

        :return:
        """
        input_shape = self.input_shape or ('?', '?')

        return f'Resize(from=({input_shape[1]}, {input_shape[0]}), to=({self.width}, {self.height}), pixformat={self.pixformat})'

    def __call__(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        return self.transform(*args, **kwargs)

    def transform(self, image, *args, **kwargs):
        """

        :param image:
        :return:
        """
        if len(image.shape) == 2 or image.shape[2] == 1:
            self.pixformat = 'gray'
        else:
            assert self.pixformat != 'gray', 'pixformat is gray, but image has 3 components'

        self.input_shape = image.shape

        return resize(image, (self.height, self.width))

    def get_template_data(self) -> dict:
        """

        :return: dict
        """
        return {
            'source_width': self.input_shape[1],
            'source_height': self.input_shape[0],
            'dest_width': self.width,
            'dest_height': self.height,
            'pixformat': self.pixformat
        }



import os.path
from pathlib import Path
from jinja2 import FileSystemLoader as Base, TemplateNotFound


class FileSystemLoader(Base):
    """
    Custom loader compatible with Windows and Unix
    """
    def __init__(self, template_folder, language, dialect=None, **kwargs):
        """
        Constructor
        :param template_folder: str
        :param language: str
        :param dialect: str
        :param kwargs:
        """
        base_folder = Path(__file__).absolute().parent.parent.parent
        super(FileSystemLoader, self).__init__(base_folder, **kwargs)
        self.template_folder = template_folder.replace('everywhereml/', 'templates/')
        self.language = language
        self.dialect = dialect

    def get_source(self, environment, template):
        """
        Return the most specific source file
        :param environment:
        :param template:
        :return:
        """
        dialect = self.dialect or "dialect"
        template = template.replace(os.path.sep, '/')
        folder = self.template_folder.replace(f'{template}/', '')
        attempts = []

        while len(folder):
            attempts += [
                f'{folder}/{self.language}/{dialect}/{template}.{self.language}.{dialect}.jinja',
                f'{folder}/{self.language}/{dialect}/{template}.{self.language}.jinja',
                f'{folder}/{self.language}/{dialect}/{template}.jinja',
                f'{folder}/{self.language}/{template}.{self.language}.{dialect}.jinja',
                f'{folder}/{self.language}/{template}.{self.language}.jinja',
                f'{folder}/{self.language}/{template}.jinja',
                f'{folder}/{template}.{self.language}.{dialect}.jinja',
                f'{folder}/{template}.{self.language}.jinja',
                f'{folder}/{template}.jinja',
            ]
            folder = os.path.dirname(folder)

        for source in attempts:
            try:
                return super().get_source(environment, source)
            except TemplateNotFound:
                pass

        raise TemplateNotFound(f'{template} in {attempts}')



import sys


class BaseCollector:
    """
    Base class for data collectors
    """
    def ask_target_name(self):
        """
        Ask user for target name
        :return:
        """
        print('This is an interactive data capturing procedure.')
        print('Keep in mind that as soon as you will enter a class name, the capturing will start, so be ready!')

        while True:
            try:
                target_name = input('Which class are you going to capture? (leave empty to exit) ').strip()

                if len(target_name) == 0 and input('Are you sure you want to exit? (y|n) ').strip().lower() == 'y':
                    break

                if len(target_name) == 0:
                    continue

                yield target_name

            except Exception as ex:
                print(ex, file=sys.stderr)


from everywhereml.code_generators.GeneratesCode import GeneratesCode


class BaseObjectDetectionPipeline(GeneratesCode):
    """
    Base class for object detection pipelines
    """
    def __init__(self, transforms: list = None, name: str = 'ObjectDetectionPipeline', **kwargs):
        """
        Constructor
        :param name: str
        :param transforms: list
        """
        self.name = name
        self.dataset = None
        self.transformed_dataset = None
        self.transforms = transforms or []
        self.num_inputs = 0
        self.num_outputs = 0
        self.input_shape = None
        self.transformed_shape = None

    def __repr__(self):
        """
        Convert to string
        :return:
        """
        return str(self)

    def __str__(self):
        """
        Convert to string
        :return:
        """
        transforms = '\n'.join([' - %s' % str(transform) for transform in self.transforms])
        steps = '\n'.join([' > %s' % str(step) for step in self.steps])

        return f'ImagePipeline: {self.name}\n---------\n{transforms}\n{steps}'

    def fit(self, image_dataset):
        """
        Fit pipeline
        :param image_dataset:
        :return:
        """
        self.input_shape = image_dataset.images[0].shape
        self.transformed_dataset = image_dataset = self.preprocess(image_dataset)
        self.num_inputs = image_dataset.num_inputs
        self.transformed_shape = self.transformed_dataset.images[0].shape

        for step in self.steps:
            step.fit(image_dataset)
            image_dataset = step.transform(image_dataset)

        self.dataset = image_dataset
        self.num_outputs = image_dataset.num_inputs

        return self

    def transform(self, image_dataset):
        """
        Transform dataset
        :param image_dataset:
        :return:
        """
        assert self.dataset is not None, 'Unfitted'
        image_dataset = self.preprocess(image_dataset)

        for step in self.steps:
            image_dataset = step.transform(image_dataset)

        self.dataset = image_dataset

        return image_dataset

    def fit_transform(self, image_dataset):
        """
        Fit then transform image_dataset
        :param image_dataset:
        :return:
        """
        return self.fit(image_dataset).dataset

    def preprocess(self, image_dataset):
        """
        Apply pre-processing transforms
        :param image_dataset:
        :return:
        """
        for transform in self.transforms:
            image_dataset = image_dataset.map(transform)

        return image_dataset

    def get_template_data(self):
        """

        :return: dict
        """
        return {
            'name': self.name,
            'transforms': self.transforms,
            'steps': self.steps,
            'num_inputs': self.num_inputs,
            'num_outputs': self.num_outputs,
            'source_width': self.input_shape[1],
            'source_height': self.input_shape[0],
            'transformed_shape': self.transformed_shape
        }

    def get_template_data_cpp(self, dialect=None):
        """

        :return: dict
        """
        working_size = max([step.num_outputs for step in self.steps])

        return {
            'working_size': working_size
        }
